@@ -7,8 +7,10 @@
 
 #include <mgba/core/config.h>
 #include <mgba/core/thread.h>
+#include <mgba-util/vfs.h>
 
 #define MAX_CATEGORY 64
+#define MAX_LOG_BUF 1024
 
 static struct mLogger* _defaultLogger = NULL;
 
@@ -80,6 +82,14 @@ void mLog(int category, enum mLogLevel level, const char* format, ...) {
 	va_end(args);
 }
 
+void mLogExplicit(struct mLogger* context, int category, enum mLogLevel level, const char* format, ...) {
+	va_list args;
+	va_start(args, format);
+	if (!context->filter || mLogFilterTest(context->filter, category, level)) {
+		context->log(context, category, level, format, args);
+	}
+}
+
 void mLogFilterInit(struct mLogFilter* filter) {
 	HashTableInit(&filter->categories, 8, NULL);
 	TableInit(&filter->levels, 8, NULL);
@@ -175,4 +185,63 @@ int mLogFilterLevels(const struct mLogFilter* filter , int category) {
 	return value;
 }
 
+void _mCoreStandardLog(struct mLogger* logger, int category, enum mLogLevel level, const char* format, va_list args) {
+	struct mStandardLogger* stdlog = (struct mStandardLogger*) logger;
+
+	if (!mLogFilterTest(logger->filter, category, level)) {
+		return;
+	}
+
+	char buffer[MAX_LOG_BUF];
+
+	// Prepare the string
+	size_t length = snprintf(buffer, sizeof(buffer), "%s: ", mLogCategoryName(category));
+	if (length < sizeof(buffer)) {
+		length += vsnprintf(buffer + length, sizeof(buffer) - length, format, args);
+	}
+	if (length < sizeof(buffer)) {
+		length += snprintf(buffer + length, sizeof(buffer) - length, "\n");
+	}
+
+	// Make sure the length doesn't exceed the size of the buffer when actually writing
+	if (length > sizeof(buffer)) {
+		length = sizeof(buffer);
+	}
+
+	if (stdlog->logToStdout) {
+		printf("%s", buffer);
+	}
+
+	if (stdlog->logFile) {
+		stdlog->logFile->write(stdlog->logFile, buffer, length);
+	}
+}
+
+void mStandardLoggerInit(struct mStandardLogger* logger) {
+	logger->d.log = _mCoreStandardLog;
+	logger->d.filter = malloc(sizeof(struct mLogFilter));
+	mLogFilterInit(logger->d.filter);
+}
+
+void mStandardLoggerDeinit(struct mStandardLogger* logger) {
+	if (logger->d.filter) {
+		mLogFilterDeinit(logger->d.filter);
+		free(logger->d.filter);
+		logger->d.filter = NULL;
+	}
+}
+
+void mStandardLoggerConfig(struct mStandardLogger* logger, struct mCoreConfig* config) {
+	bool logToFile = false;
+	const char* logFile = mCoreConfigGetValue(config, "logFile");
+	mCoreConfigGetBoolValue(config, "logToStdout", &logger->logToStdout);
+	mCoreConfigGetBoolValue(config, "logToFile", &logToFile);
+
+	if (logToFile && logFile) {
+		logger->logFile = VFileOpen(logFile, O_WRONLY | O_CREAT | O_APPEND);
+	}
+
+	mLogFilterLoad(logger->d.filter, config);
+}
+
 mLOG_DEFINE_CATEGORY(STATUS, "Status", "core.status")
