@@ -11,8 +11,11 @@
 #include <mgba/internal/gb/gb.h>
 #include <mgba/internal/gb/serialize.h>
 #include <mgba/internal/gb/io.h>
+#ifdef M_CORE_GBA
+#include <mgba/internal/gba/audio.h>
+#endif
 
-#ifdef _3DS
+#ifdef __3DS__
 #define blip_add_delta blip_add_delta_fast
 #endif
 
@@ -21,6 +24,8 @@
 const uint32_t DMG_SM83_FREQUENCY = 0x400000;
 static const int CLOCKS_PER_BLIP_FRAME = 0x1000;
 static const unsigned BLIP_BUFFER_SIZE = 0x4000;
+static const int SAMPLE_INTERVAL = 32;
+static const int FILTER = 65368;
 const int GB_AUDIO_VOLUME_MAX = 0x100;
 
 static bool _writeSweep(struct GBAudioSweep* sweep, uint8_t value);
@@ -35,17 +40,21 @@ static void _updateEnvelopeDead(struct GBAudioEnvelope* envelope);
 static bool _updateSweep(struct GBAudioSquareChannel* sweep, bool initial);
 
 static void _updateSquareSample(struct GBAudioSquareChannel* ch);
-static int32_t _updateSquareChannel(struct GBAudioSquareChannel* ch);
 
 static int16_t _coalesceNoiseChannel(struct GBAudioNoiseChannel* ch);
 
 static void _updateFrame(struct mTiming* timing, void* user, uint32_t cyclesLate);
-static void _updateChannel1(struct mTiming* timing, void* user, uint32_t cyclesLate);
-static void _updateChannel2(struct mTiming* timing, void* user, uint32_t cyclesLate);
-static void _updateChannel3(struct mTiming* timing, void* user, uint32_t cyclesLate);
-static void _fadeChannel3(struct mTiming* timing, void* user, uint32_t cyclesLate);
 static void _sample(struct mTiming* timing, void* user, uint32_t cyclesLate);
 
+static void GBAudioSample(struct GBAudio* audio, int32_t timestamp);
+
+static const int _squareChannelDuty[4][8] = {
+	{ 0, 0, 0, 0, 0, 0, 0, 1 },
+	{ 1, 0, 0, 0, 0, 0, 0, 1 },
+	{ 1, 0, 0, 0, 0, 1, 1, 1 },
+	{ 0, 1, 1, 1, 1, 1, 1, 0 },
+};
+
 void GBAudioInit(struct GBAudio* audio, size_t samples, uint8_t* nr52, enum GBAudioStyle style) {
 	audio->samples = samples;
 	audio->left = blip_new(BLIP_BUFFER_SIZE);
@@ -67,30 +76,9 @@ void GBAudioInit(struct GBAudio* audio, size_t samples, uint8_t* nr52, enum GBAu
 		audio->timingFactor = 2;
 	}
 
-	audio->frameEvent.context = audio;
 	audio->frameEvent.name = "GB Audio Frame Sequencer";
 	audio->frameEvent.callback = _updateFrame;
 	audio->frameEvent.priority = 0x10;
-	audio->ch1Event.context = audio;
-	audio->ch1Event.name = "GB Audio Channel 1";
-	audio->ch1Event.callback = _updateChannel1;
-	audio->ch1Event.priority = 0x11;
-	audio->ch2Event.context = audio;
-	audio->ch2Event.name = "GB Audio Channel 2";
-	audio->ch2Event.callback = _updateChannel2;
-	audio->ch2Event.priority = 0x12;
-	audio->ch3Event.context = audio;
-	audio->ch3Event.name = "GB Audio Channel 3";
-	audio->ch3Event.callback = _updateChannel3;
-	audio->ch3Event.priority = 0x13;
-	audio->ch3Fade.context = audio;
-	audio->ch3Fade.name = "GB Audio Channel 3 Memory";
-	audio->ch3Fade.callback = _fadeChannel3;
-	audio->ch3Fade.priority = 0x14;
-	audio->ch4Event.context = audio;
-	audio->ch4Event.name = "GB Audio Channel 4";
-	audio->ch4Event.callback = NULL; // This is pending removal, so calling it will crash
-	audio->ch4Event.priority = 0x15;
 	audio->sampleEvent.context = audio;
 	audio->sampleEvent.name = "GB Audio Sample";
 	audio->sampleEvent.callback = _sample;
@@ -103,19 +91,10 @@ void GBAudioDeinit(struct GBAudio* audio) {
 }
 
 void GBAudioReset(struct GBAudio* audio) {
-	mTimingDeschedule(audio->timing, &audio->frameEvent);
-	mTimingDeschedule(audio->timing, &audio->ch1Event);
-	mTimingDeschedule(audio->timing, &audio->ch2Event);
-	mTimingDeschedule(audio->timing, &audio->ch3Event);
-	mTimingDeschedule(audio->timing, &audio->ch3Fade);
-	mTimingDeschedule(audio->timing, &audio->ch4Event);
 	mTimingDeschedule(audio->timing, &audio->sampleEvent);
 	if (audio->style != GB_AUDIO_GBA) {
 		mTimingSchedule(audio->timing, &audio->sampleEvent, 0);
 	}
-	if (audio->style == GB_AUDIO_GBA) {
-		mTimingSchedule(audio->timing, &audio->frameEvent, 0);
-	}
 	audio->ch1 = (struct GBAudioSquareChannel) { .sweep = { .time = 8 }, .envelope = { .dead = 2 } };
 	audio->ch2 = (struct GBAudioSquareChannel) { .envelope = { .dead = 2 } };
 	audio->ch3 = (struct GBAudioWaveChannel) { .bank = 0 };
@@ -139,7 +118,9 @@ void GBAudioReset(struct GBAudio* audio) {
 	audio->ch3.wavedata8[15] = 0xFF;
 	audio->ch4 = (struct GBAudioNoiseChannel) { .envelope = { .dead = 2 } };
 	audio->frame = 0;
-	audio->sampleInterval = 128;
+	audio->sampleInterval = SAMPLE_INTERVAL * GB_MAX_SAMPLES;
+	audio->lastSample = 0;
+	audio->sampleIndex = 0;
 	audio->lastLeft = 0;
 	audio->lastRight = 0;
 	audio->capLeft = 0;
@@ -166,32 +147,35 @@ void GBAudioResizeBuffer(struct GBAudio* audio, size_t samples) {
 }
 
 void GBAudioWriteNR10(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x1);
 	if (!_writeSweep(&audio->ch1.sweep, value)) {
-		mTimingDeschedule(audio->timing, &audio->ch1Event);
 		audio->playingCh1 = false;
 		*audio->nr52 &= ~0x0001;
 	}
 }
 
 void GBAudioWriteNR11(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x1);
 	_writeDuty(&audio->ch1.envelope, value);
 	audio->ch1.control.length = 64 - audio->ch1.envelope.length;
 }
 
 void GBAudioWriteNR12(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x1);
 	if (!_writeEnvelope(&audio->ch1.envelope, value, audio->style)) {
-		mTimingDeschedule(audio->timing, &audio->ch1Event);
 		audio->playingCh1 = false;
 		*audio->nr52 &= ~0x0001;
 	}
 }
 
 void GBAudioWriteNR13(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x1);
 	audio->ch1.control.frequency &= 0x700;
 	audio->ch1.control.frequency |= GBAudioRegisterControlGetFrequency(value);
 }
 
 void GBAudioWriteNR14(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x1);
 	audio->ch1.control.frequency &= 0xFF;
 	audio->ch1.control.frequency |= GBAudioRegisterControlGetFrequency(value << 8);
 	bool wasStop = audio->ch1.control.stop;
@@ -199,7 +183,6 @@ void GBAudioWriteNR14(struct GBAudio* audio, uint8_t value) {
 	if (!wasStop && audio->ch1.control.stop && audio->ch1.control.length && !(audio->frame & 1)) {
 		--audio->ch1.control.length;
 		if (audio->ch1.control.length == 0) {
-			mTimingDeschedule(audio->timing, &audio->ch1Event);
 			audio->playingCh1 = false;
 		}
 	}
@@ -216,35 +199,34 @@ void GBAudioWriteNR14(struct GBAudio* audio, uint8_t value) {
 				--audio->ch1.control.length;
 			}
 		}
-		if (audio->playingCh1 && audio->ch1.envelope.dead != 2) {
-			_updateSquareChannel(&audio->ch1);
-			mTimingDeschedule(audio->timing, &audio->ch1Event);
-			mTimingSchedule(audio->timing, &audio->ch1Event, 0);
-		}
+		_updateSquareSample(&audio->ch1);
 	}
 	*audio->nr52 &= ~0x0001;
 	*audio->nr52 |= audio->playingCh1;
 }
 
 void GBAudioWriteNR21(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x2);
 	_writeDuty(&audio->ch2.envelope, value);
 	audio->ch2.control.length = 64 - audio->ch2.envelope.length;
 }
 
 void GBAudioWriteNR22(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x2);
 	if (!_writeEnvelope(&audio->ch2.envelope, value, audio->style)) {
-		mTimingDeschedule(audio->timing, &audio->ch2Event);
 		audio->playingCh2 = false;
 		*audio->nr52 &= ~0x0002;
 	}
 }
 
 void GBAudioWriteNR23(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x2);
 	audio->ch2.control.frequency &= 0x700;
 	audio->ch2.control.frequency |= GBAudioRegisterControlGetFrequency(value);
 }
 
 void GBAudioWriteNR24(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x2);
 	audio->ch2.control.frequency &= 0xFF;
 	audio->ch2.control.frequency |= GBAudioRegisterControlGetFrequency(value << 8);
 	bool wasStop = audio->ch2.control.stop;
@@ -252,7 +234,6 @@ void GBAudioWriteNR24(struct GBAudio* audio, uint8_t value) {
 	if (!wasStop && audio->ch2.control.stop && audio->ch2.control.length && !(audio->frame & 1)) {
 		--audio->ch2.control.length;
 		if (audio->ch2.control.length == 0) {
-			mTimingDeschedule(audio->timing, &audio->ch2Event);
 			audio->playingCh2 = false;
 		}
 	}
@@ -265,39 +246,39 @@ void GBAudioWriteNR24(struct GBAudio* audio, uint8_t value) {
 				--audio->ch2.control.length;
 			}
 		}
-		if (audio->playingCh2 && audio->ch2.envelope.dead != 2) {
-			_updateSquareChannel(&audio->ch2);
-			mTimingDeschedule(audio->timing, &audio->ch2Event);
-			mTimingSchedule(audio->timing, &audio->ch2Event, 0);
-		}
+		_updateSquareSample(&audio->ch2);
 	}
 	*audio->nr52 &= ~0x0002;
 	*audio->nr52 |= audio->playingCh2 << 1;
 }
 
 void GBAudioWriteNR30(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x4);
 	audio->ch3.enable = GBAudioRegisterBankGetEnable(value);
 	if (!audio->ch3.enable) {
-		mTimingDeschedule(audio->timing, &audio->ch3Event);
 		audio->playingCh3 = false;
 		*audio->nr52 &= ~0x0004;
 	}
 }
 
 void GBAudioWriteNR31(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x4);
 	audio->ch3.length = 256 - value;
 }
 
 void GBAudioWriteNR32(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x4);
 	audio->ch3.volume = GBAudioRegisterBankVolumeGetVolumeGB(value);
 }
 
 void GBAudioWriteNR33(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x4);
 	audio->ch3.rate &= 0x700;
 	audio->ch3.rate |= GBAudioRegisterControlGetRate(value);
 }
 
 void GBAudioWriteNR34(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x4);
 	audio->ch3.rate &= 0xFF;
 	audio->ch3.rate |= GBAudioRegisterControlGetRate(value << 8);
 	bool wasStop = audio->ch3.stop;
@@ -333,25 +314,23 @@ void GBAudioWriteNR34(struct GBAudio* audio, uint8_t value) {
 			audio->ch3.sample = 0;
 		}
 	}
-	mTimingDeschedule(audio->timing, &audio->ch3Fade);
-	mTimingDeschedule(audio->timing, &audio->ch3Event);
 	if (audio->playingCh3) {
 		audio->ch3.readable = audio->style != GB_AUDIO_DMG;
 		// TODO: Where does this cycle delay come from?
-		mTimingSchedule(audio->timing, &audio->ch3Event, audio->timingFactor * (4 + 2 * (2048 - audio->ch3.rate)));
+		audio->ch3.nextUpdate = mTimingCurrentTime(audio->timing) + (6 + 2 * (2048 - audio->ch3.rate)) * audio->timingFactor;
 	}
 	*audio->nr52 &= ~0x0004;
 	*audio->nr52 |= audio->playingCh3 << 2;
 }
 
 void GBAudioWriteNR41(struct GBAudio* audio, uint8_t value) {
-	GBAudioUpdateChannel4(audio);
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x8);
 	_writeDuty(&audio->ch4.envelope, value);
 	audio->ch4.length = 64 - audio->ch4.envelope.length;
 }
 
 void GBAudioWriteNR42(struct GBAudio* audio, uint8_t value) {
-	GBAudioUpdateChannel4(audio);
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x8);
 	if (!_writeEnvelope(&audio->ch4.envelope, value, audio->style)) {
 		audio->playingCh4 = false;
 		*audio->nr52 &= ~0x0008;
@@ -359,14 +338,14 @@ void GBAudioWriteNR42(struct GBAudio* audio, uint8_t value) {
 }
 
 void GBAudioWriteNR43(struct GBAudio* audio, uint8_t value) {
-	GBAudioUpdateChannel4(audio);
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x8);
 	audio->ch4.ratio = GBAudioRegisterNoiseFeedbackGetRatio(value);
 	audio->ch4.frequency = GBAudioRegisterNoiseFeedbackGetFrequency(value);
 	audio->ch4.power = GBAudioRegisterNoiseFeedbackGetPower(value);
 }
 
 void GBAudioWriteNR44(struct GBAudio* audio, uint8_t value) {
-	GBAudioUpdateChannel4(audio);
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x8);
 	bool wasStop = audio->ch4.stop;
 	audio->ch4.stop = GBAudioRegisterNoiseControlGetStop(value);
 	if (!wasStop && audio->ch4.stop && audio->ch4.length && !(audio->frame & 1)) {
@@ -389,7 +368,7 @@ void GBAudioWriteNR44(struct GBAudio* audio, uint8_t value) {
 				--audio->ch4.length;
 			}
 		}
-		if (audio->playingCh4 && audio->ch4.envelope.dead != 2) {
+		if (audio->playingCh4) {
 			audio->ch4.lastEvent = mTimingCurrentTime(audio->timing);
 		}
 	}
@@ -398,11 +377,13 @@ void GBAudioWriteNR44(struct GBAudio* audio, uint8_t value) {
 }
 
 void GBAudioWriteNR50(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x2);
 	audio->volumeRight = GBRegisterNR50GetVolumeRight(value);
 	audio->volumeLeft = GBRegisterNR50GetVolumeLeft(value);
 }
 
 void GBAudioWriteNR51(struct GBAudio* audio, uint8_t value) {
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x2);
 	audio->ch1Right = GBRegisterNR51GetCh1Right(value);
 	audio->ch2Right = GBRegisterNR51GetCh2Right(value);
 	audio->ch3Right = GBRegisterNR51GetCh3Right(value);
@@ -476,17 +457,148 @@ void GBAudioWriteNR52(struct GBAudio* audio, uint8_t value) {
 		audio->skipFrame = false;
 		audio->frame = 7;
 
-		if (audio->p && audio->p->timer.internalDiv & 0x400) {
+		if (audio->p && audio->p->timer.internalDiv & (0x100 << audio->p->doubleSpeed)) {
 			audio->skipFrame = true;
 		}
 	}
 }
 
 void _updateFrame(struct mTiming* timing, void* user, uint32_t cyclesLate) {
-	struct GBAudio* audio = user;
-	GBAudioUpdateFrame(audio);
-	if (audio->style == GB_AUDIO_GBA) {
-		mTimingSchedule(timing, &audio->frameEvent, audio->timingFactor * FRAME_CYCLES - cyclesLate);
+#ifdef M_CORE_GBA
+	struct GBAAudio* audio = user;
+	GBAAudioSample(audio, mTimingCurrentTime(timing));
+	mTimingSchedule(timing, &audio->psg.frameEvent, audio->psg.timingFactor * FRAME_CYCLES - cyclesLate);
+	GBAudioUpdateFrame(&audio->psg);
+#endif
+}
+
+void GBAudioRun(struct GBAudio* audio, int32_t timestamp, int channels) {
+	if (!audio->enable) {
+		return;
+	}
+	if (audio->p && channels != 0xF && timestamp - audio->lastSample > (int) (SAMPLE_INTERVAL * audio->timingFactor)) {
+		GBAudioSample(audio, timestamp);
+	}
+
+	if (audio->playingCh1 && (channels & 0x1)) {
+		int period = 4 * (2048 - audio->ch1.control.frequency) * audio->timingFactor;
+		int32_t diff = timestamp - audio->ch1.lastUpdate;
+		if (diff >= period) {
+			diff /= period;
+			audio->ch1.index = (audio->ch1.index + diff) & 7;
+			audio->ch1.lastUpdate += diff * period;
+			_updateSquareSample(&audio->ch1);
+		}
+	}
+	if (audio->playingCh2 && (channels & 0x2)) {
+		int period = 4 * (2048 - audio->ch2.control.frequency) * audio->timingFactor;
+		int32_t diff = timestamp - audio->ch2.lastUpdate;
+		if (diff >= period) {
+			diff /= period;
+			audio->ch2.index = (audio->ch2.index + diff) & 7;
+			audio->ch2.lastUpdate += diff * period;
+			_updateSquareSample(&audio->ch2);
+		}
+	}
+	if (audio->playingCh3 && (channels & 0x4)) {
+		int cycles = 2 * (2048 - audio->ch3.rate) * audio->timingFactor;
+		int32_t diff = timestamp - audio->ch3.nextUpdate;
+		if (diff >= 0) {
+			diff = (diff / cycles) + 1;
+			int volume;
+			switch (audio->ch3.volume) {
+			case 0:
+				volume = 4;
+				break;
+			case 1:
+				volume = 0;
+				break;
+			case 2:
+				volume = 1;
+				break;
+			default:
+			case 3:
+				volume = 2;
+				break;
+			}
+			int start = 7;
+			int end = 0;
+			int mask = 0x1F;
+			int iter;
+			switch (audio->style) {
+			case GB_AUDIO_DMG:
+			default:
+				audio->ch3.window += diff;
+				audio->ch3.window &= 0x1F;
+				audio->ch3.sample = audio->ch3.wavedata8[audio->ch3.window >> 1];
+				if (!(audio->ch3.window & 1)) {
+					audio->ch3.sample >>= 4;
+				}
+				audio->ch3.sample &= 0xF;
+				break;
+			case GB_AUDIO_GBA:
+				if (audio->ch3.size) {
+					mask = 0x3F;
+				} else if (audio->ch3.bank) {
+					end = 4;
+				} else {
+					start = 3;
+				}
+				for (iter = 0; iter < (diff & mask); ++iter) {
+					uint32_t bitsCarry = audio->ch3.wavedata32[end] & 0x000000F0;
+					uint32_t bits;
+					int i;
+					for (i = start; i >= end; --i) {
+						bits = audio->ch3.wavedata32[i] & 0x000000F0;
+						audio->ch3.wavedata32[i] = ((audio->ch3.wavedata32[i] & 0x0F0F0F0F) << 4) | ((audio->ch3.wavedata32[i] & 0xF0F0F000) >> 12);
+						audio->ch3.wavedata32[i] |= bitsCarry << 20;
+						bitsCarry = bits;
+					}
+					audio->ch3.sample = bitsCarry >> 4;
+				}
+				break;
+			}
+			if (audio->ch3.volume > 3) {
+				audio->ch3.sample += audio->ch3.sample << 1;
+			}
+			audio->ch3.sample >>= volume;
+			audio->ch3.nextUpdate += diff * cycles;
+			audio->ch3.readable = true;
+		}
+		if (audio->style == GB_AUDIO_DMG && audio->ch3.readable) {
+			diff = timestamp - audio->ch3.nextUpdate + cycles;
+			if (diff >= 4) {
+				audio->ch3.readable = false;
+			}
+		}
+	}
+	if (audio->playingCh4 && (channels & 0x8)) {
+		int32_t cycles = audio->ch4.ratio ? 2 * audio->ch4.ratio : 1;
+		cycles <<= audio->ch4.frequency;
+		cycles *= 8 * audio->timingFactor;
+
+		int32_t diff = timestamp - audio->ch4.lastEvent;
+		if (diff >= cycles) {
+			int32_t last;
+			int samples = 0;
+			int positiveSamples = 0;
+			int lsb;
+			int coeff = 0x60;
+			if (!audio->ch4.power) {
+				coeff <<= 8;
+			}
+			for (last = 0; last + cycles <= diff; last += cycles) {
+				lsb = audio->ch4.lfsr & 1;
+				audio->ch4.lfsr >>= 1;
+				audio->ch4.lfsr ^= lsb * coeff;
+				++samples;
+				positiveSamples += lsb;
+			}
+			audio->ch4.sample = lsb * audio->ch4.envelope.currentVolume;
+			audio->ch4.nSamples += samples;
+			audio->ch4.samples += positiveSamples * audio->ch4.envelope.currentVolume;
+			audio->ch4.lastEvent += last;
+		}
 	}
 }
 
@@ -498,6 +610,8 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 		audio->skipFrame = false;
 		return;
 	}
+	GBAudioRun(audio, mTimingCurrentTime(audio->timing), 0x7);
+
 	int frame = (audio->frame + 1) & 7;
 	audio->frame = frame;
 
@@ -510,9 +624,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 				audio->playingCh1 = _updateSweep(&audio->ch1, false);
 				*audio->nr52 &= ~0x0001;
 				*audio->nr52 |= audio->playingCh1;
-				if (!audio->playingCh1) {
-					mTimingDeschedule(audio->timing, &audio->ch1Event);
-				}
 			}
 		}
 		// Fall through
@@ -521,7 +632,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 		if (audio->ch1.control.length && audio->ch1.control.stop) {
 			--audio->ch1.control.length;
 			if (audio->ch1.control.length == 0) {
-				mTimingDeschedule(audio->timing, &audio->ch1Event);
 				audio->playingCh1 = 0;
 				*audio->nr52 &= ~0x0001;
 			}
@@ -530,7 +640,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 		if (audio->ch2.control.length && audio->ch2.control.stop) {
 			--audio->ch2.control.length;
 			if (audio->ch2.control.length == 0) {
-				mTimingDeschedule(audio->timing, &audio->ch2Event);
 				audio->playingCh2 = 0;
 				*audio->nr52 &= ~0x0002;
 			}
@@ -539,7 +648,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 		if (audio->ch3.length && audio->ch3.stop) {
 			--audio->ch3.length;
 			if (audio->ch3.length == 0) {
-				mTimingDeschedule(audio->timing, &audio->ch3Event);
 				audio->playingCh3 = 0;
 				*audio->nr52 &= ~0x0004;
 			}
@@ -548,7 +656,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 		if (audio->ch4.length && audio->ch4.stop) {
 			--audio->ch4.length;
 			if (audio->ch4.length == 0) {
-				GBAudioUpdateChannel4(audio);
 				audio->playingCh4 = 0;
 				*audio->nr52 &= ~0x0008;
 			}
@@ -559,9 +666,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 			--audio->ch1.envelope.nextStep;
 			if (audio->ch1.envelope.nextStep == 0) {
 				_updateEnvelope(&audio->ch1.envelope);
-				if (audio->ch1.envelope.dead == 2) {
-					mTimingDeschedule(audio->timing, &audio->ch1Event);
-				}
 				_updateSquareSample(&audio->ch1);
 			}
 		}
@@ -570,9 +674,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 			--audio->ch2.envelope.nextStep;
 			if (audio->ch2.envelope.nextStep == 0) {
 				_updateEnvelope(&audio->ch2.envelope);
-				if (audio->ch2.envelope.dead == 2) {
-					mTimingDeschedule(audio->timing, &audio->ch2Event);
-				}
 				_updateSquareSample(&audio->ch2);
 			}
 		}
@@ -580,7 +681,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 		if (audio->playingCh4 && !audio->ch4.envelope.dead) {
 			--audio->ch4.envelope.nextStep;
 			if (audio->ch4.envelope.nextStep == 0) {
-				GBAudioUpdateChannel4(audio);
 				int8_t sample = audio->ch4.sample;
 				_updateEnvelope(&audio->ch4.envelope);
 				audio->ch4.sample = (sample > 0) * audio->ch4.envelope.currentVolume;
@@ -594,31 +694,6 @@ void GBAudioUpdateFrame(struct GBAudio* audio) {
 	}
 }
 
-void GBAudioUpdateChannel4(struct GBAudio* audio) {
-	struct GBAudioNoiseChannel* ch = &audio->ch4;
-	if (ch->envelope.dead == 2 || !audio->playingCh4) {
-		return;
-	}
-
-	int32_t cycles = ch->ratio ? 2 * ch->ratio : 1;
-	cycles <<= ch->frequency;
-	cycles *= 8 * audio->timingFactor;
-
-	uint32_t last = 0;
-	uint32_t now = mTimingCurrentTime(audio->timing) - ch->lastEvent;
-
-	for (; last + cycles <= now; last += cycles) {
-		int lsb = ch->lfsr & 1;
-		ch->sample = lsb * ch->envelope.currentVolume;
-		++ch->nSamples;
-		ch->samples += ch->sample;
-		ch->lfsr >>= 1;
-		ch->lfsr ^= (lsb * 0x60) << (ch->power ? 0 : 8);
-	}
-
-	ch->lastEvent += last;
-}
-
 void GBAudioSamplePSG(struct GBAudio* audio, int16_t* left, int16_t* right) {
 	int dcOffset = audio->style == GB_AUDIO_GBA ? 0 : -0x8;
 	int sampleLeft = dcOffset;
@@ -658,7 +733,6 @@ void GBAudioSamplePSG(struct GBAudio* audio, int16_t* left, int16_t* right) {
 	sampleRight <<= 3;
 
 	if (!audio->forceDisableCh[3]) {
-		GBAudioUpdateChannel4(audio);
 		int16_t sample = audio->style == GB_AUDIO_GBA ? (audio->ch4.sample << 3) : _coalesceNoiseChannel(&audio->ch4);
 		if (audio->ch4Left) {
 			sampleLeft += sample;
@@ -673,40 +747,64 @@ void GBAudioSamplePSG(struct GBAudio* audio, int16_t* left, int16_t* right) {
 	*right = sampleRight * (1 + audio->volumeRight);
 }
 
+void GBAudioSample(struct GBAudio* audio, int32_t timestamp) {
+	int interval = SAMPLE_INTERVAL * audio->timingFactor;
+	timestamp -= audio->lastSample;
+	timestamp -= audio->sampleIndex * interval;
+
+	int sample;
+	for (sample = audio->sampleIndex; timestamp >= interval && sample < GB_MAX_SAMPLES; ++sample, timestamp -= interval) {
+		int16_t sampleLeft = 0;
+		int16_t sampleRight = 0;
+		GBAudioRun(audio, sample * interval + audio->lastSample, 0xF);
+		GBAudioSamplePSG(audio, &sampleLeft, &sampleRight);
+		sampleLeft = (sampleLeft * audio->masterVolume * 6) >> 7;
+		sampleRight = (sampleRight * audio->masterVolume * 6) >> 7;
+
+		int16_t degradedLeft = sampleLeft - (audio->capLeft >> 16);
+		int16_t degradedRight = sampleRight - (audio->capRight >> 16);
+		audio->capLeft = (sampleLeft << 16) - degradedLeft * FILTER;
+		audio->capRight = (sampleRight << 16) - degradedRight * FILTER;
+
+		audio->currentSamples[sample].left = degradedLeft;
+		audio->currentSamples[sample].right = degradedRight;
+	}
+
+	audio->sampleIndex = sample;
+	if (sample == GB_MAX_SAMPLES) {
+		audio->lastSample += interval * GB_MAX_SAMPLES;
+		audio->sampleIndex = 0;
+	}
+}
+
 static void _sample(struct mTiming* timing, void* user, uint32_t cyclesLate) {
 	struct GBAudio* audio = user;
-	int16_t sampleLeft = 0;
-	int16_t sampleRight = 0;
-	GBAudioSamplePSG(audio, &sampleLeft, &sampleRight);
-	sampleLeft = (sampleLeft * audio->masterVolume * 6) >> 7;
-	sampleRight = (sampleRight * audio->masterVolume * 6) >> 7;
+	GBAudioSample(audio, mTimingCurrentTime(audio->timing));
 
 	mCoreSyncLockAudio(audio->p->sync);
 	unsigned produced;
-
-	int16_t degradedLeft = sampleLeft - (audio->capLeft >> 16);
-	int16_t degradedRight = sampleRight - (audio->capRight >> 16);
-	audio->capLeft = (sampleLeft << 16) - degradedLeft * 65184;
-	audio->capRight = (sampleRight << 16) - degradedRight * 65184;
-	sampleLeft = degradedLeft;
-	sampleRight = degradedRight;
-
-	if ((size_t) blip_samples_avail(audio->left) < audio->samples) {
-		blip_add_delta(audio->left, audio->clock, sampleLeft - audio->lastLeft);
-		blip_add_delta(audio->right, audio->clock, sampleRight - audio->lastRight);
-		audio->lastLeft = sampleLeft;
-		audio->lastRight = sampleRight;
-		audio->clock += audio->sampleInterval;
-		if (audio->clock >= CLOCKS_PER_BLIP_FRAME) {
-			blip_end_frame(audio->left, CLOCKS_PER_BLIP_FRAME);
-			blip_end_frame(audio->right, CLOCKS_PER_BLIP_FRAME);
-			audio->clock -= CLOCKS_PER_BLIP_FRAME;
+	int i;
+	for (i = 0; i < GB_MAX_SAMPLES; ++i) {
+		int16_t sampleLeft = audio->currentSamples[i].left;
+		int16_t sampleRight = audio->currentSamples[i].right;
+		if ((size_t) blip_samples_avail(audio->left) < audio->samples) {
+			blip_add_delta(audio->left, audio->clock, sampleLeft - audio->lastLeft);
+			blip_add_delta(audio->right, audio->clock, sampleRight - audio->lastRight);
+			audio->lastLeft = sampleLeft;
+			audio->lastRight = sampleRight;
+			audio->clock += SAMPLE_INTERVAL;
+			if (audio->clock >= CLOCKS_PER_BLIP_FRAME) {
+				blip_end_frame(audio->left, CLOCKS_PER_BLIP_FRAME);
+				blip_end_frame(audio->right, CLOCKS_PER_BLIP_FRAME);
+				audio->clock -= CLOCKS_PER_BLIP_FRAME;
+			}
+		}
+		if (audio->p->stream && audio->p->stream->postAudioFrame) {
+			audio->p->stream->postAudioFrame(audio->p->stream, sampleLeft, sampleRight);
 		}
 	}
+
 	produced = blip_samples_avail(audio->left);
-	if (audio->p->stream && audio->p->stream->postAudioFrame) {
-		audio->p->stream->postAudioFrame(audio->p->stream, sampleLeft, sampleRight);
-	}
 	bool wait = produced >= audio->samples;
 	if (!mCoreSyncProduceAudio(audio->p->sync, audio->left, audio->samples)) {
 		// Interrupted
@@ -769,26 +867,7 @@ bool _writeEnvelope(struct GBAudioEnvelope* envelope, uint8_t value, enum GBAudi
 }
 
 static void _updateSquareSample(struct GBAudioSquareChannel* ch) {
-	ch->sample = ch->control.hi * ch->envelope.currentVolume;
-}
-
-static int32_t _updateSquareChannel(struct GBAudioSquareChannel* ch) {
-	ch->control.hi = !ch->control.hi;
-	_updateSquareSample(ch);
-	int period = 4 * (2048 - ch->control.frequency);
-	switch (ch->envelope.duty) {
-	case 0:
-		return ch->control.hi ? period : period * 7;
-	case 1:
-		return ch->control.hi ? period * 2 : period * 6;
-	case 2:
-		return period * 4;
-	case 3:
-		return ch->control.hi ? period * 6 : period * 2;
-	default:
-		// This should never be hit
-		return period * 4;
-	}
+	ch->sample = _squareChannelDuty[ch->envelope.duty][ch->index] * ch->envelope.currentVolume;
 }
 
 static int16_t _coalesceNoiseChannel(struct GBAudioNoiseChannel* ch) {
@@ -860,94 +939,6 @@ static bool _updateSweep(struct GBAudioSquareChannel* ch, bool initial) {
 	return true;
 }
 
-static void _updateChannel1(struct mTiming* timing, void* user, uint32_t cyclesLate) {
-	struct GBAudio* audio = user;
-	struct GBAudioSquareChannel* ch = &audio->ch1;
-	int cycles = _updateSquareChannel(ch);
-	mTimingSchedule(timing, &audio->ch1Event, audio->timingFactor * cycles - cyclesLate);
-}
-
-static void _updateChannel2(struct mTiming* timing, void* user, uint32_t cyclesLate) {
-	struct GBAudio* audio = user;
-	struct GBAudioSquareChannel* ch = &audio->ch2;
-	int cycles = _updateSquareChannel(ch);
-	mTimingSchedule(timing, &audio->ch2Event, audio->timingFactor * cycles - cyclesLate);
-}
-
-static void _updateChannel3(struct mTiming* timing, void* user, uint32_t cyclesLate) {
-	struct GBAudio* audio = user;
-	struct GBAudioWaveChannel* ch = &audio->ch3;
-	int i;
-	int volume;
-	switch (ch->volume) {
-	case 0:
-		volume = 4;
-		break;
-	case 1:
-		volume = 0;
-		break;
-	case 2:
-		volume = 1;
-		break;
-	default:
-	case 3:
-		volume = 2;
-		break;
-	}
-	int start;
-	int end;
-	switch (audio->style) {
-	case GB_AUDIO_DMG:
-	default:
-		++ch->window;
-		ch->window &= 0x1F;
-		ch->sample = ch->wavedata8[ch->window >> 1];
-		if (!(ch->window & 1)) {
-			ch->sample >>= 4;
-		}
-		ch->sample &= 0xF;
-		break;
-	case GB_AUDIO_GBA:
-		if (ch->size) {
-			start = 7;
-			end = 0;
-		} else if (ch->bank) {
-			start = 7;
-			end = 4;
-		} else {
-			start = 3;
-			end = 0;
-		}
-		uint32_t bitsCarry = ch->wavedata32[end] & 0x000000F0;
-		uint32_t bits;
-		for (i = start; i >= end; --i) {
-			bits = ch->wavedata32[i] & 0x000000F0;
-			ch->wavedata32[i] = ((ch->wavedata32[i] & 0x0F0F0F0F) << 4) | ((ch->wavedata32[i] & 0xF0F0F000) >> 12);
-			ch->wavedata32[i] |= bitsCarry << 20;
-			bitsCarry = bits;
-		}
-		ch->sample = bitsCarry >> 4;
-		break;
-	}
-	if (ch->volume > 3) {
-		ch->sample += ch->sample << 1;
-	}
-	ch->sample >>= volume;
-	audio->ch3.readable = true;
-	if (audio->style == GB_AUDIO_DMG) {
-		mTimingDeschedule(audio->timing, &audio->ch3Fade);
-		mTimingSchedule(timing, &audio->ch3Fade, 4 - cyclesLate);
-	}
-	int cycles = 2 * (2048 - ch->rate);
-	mTimingSchedule(timing, &audio->ch3Event, audio->timingFactor * cycles - cyclesLate);
-}
-static void _fadeChannel3(struct mTiming* timing, void* user, uint32_t cyclesLate) {
-	UNUSED(timing);
-	UNUSED(cyclesLate);
-	struct GBAudio* audio = user;
-	audio->ch3.readable = false;
-}
-
 void GBAudioPSGSerialize(const struct GBAudio* audio, struct GBSerializedPSGState* state, uint32_t* flagsOut) {
 	uint32_t flags = 0;
 	uint32_t sweep = 0;
@@ -961,30 +952,29 @@ void GBAudioPSGSerialize(const struct GBAudio* audio, struct GBSerializedPSGStat
 
 	flags = GBSerializedAudioFlagsSetCh1Volume(flags, audio->ch1.envelope.currentVolume);
 	flags = GBSerializedAudioFlagsSetCh1Dead(flags, audio->ch1.envelope.dead);
-	flags = GBSerializedAudioFlagsSetCh1Hi(flags, audio->ch1.control.hi);
 	flags = GBSerializedAudioFlagsSetCh1SweepEnabled(flags, audio->ch1.sweep.enable);
 	flags = GBSerializedAudioFlagsSetCh1SweepOccurred(flags, audio->ch1.sweep.occurred);
 	ch1Flags = GBSerializedAudioEnvelopeSetLength(ch1Flags, audio->ch1.control.length);
 	ch1Flags = GBSerializedAudioEnvelopeSetNextStep(ch1Flags, audio->ch1.envelope.nextStep);
 	ch1Flags = GBSerializedAudioEnvelopeSetFrequency(ch1Flags, audio->ch1.sweep.realFrequency);
+	ch1Flags = GBSerializedAudioEnvelopeSetDutyIndex(ch1Flags, audio->ch1.index);
 	sweep = GBSerializedAudioSweepSetTime(sweep, audio->ch1.sweep.time & 7);
 	STORE_32LE(ch1Flags, 0, &state->ch1.envelope);
 	STORE_32LE(sweep, 0, &state->ch1.sweep);
-	STORE_32LE(audio->ch1Event.when - mTimingCurrentTime(audio->timing), 0, &state->ch1.nextEvent);
+	STORE_32LE(audio->ch1.lastUpdate - mTimingCurrentTime(audio->timing), 0, &state->ch1.lastUpdate);
 
 	flags = GBSerializedAudioFlagsSetCh2Volume(flags, audio->ch2.envelope.currentVolume);
 	flags = GBSerializedAudioFlagsSetCh2Dead(flags, audio->ch2.envelope.dead);
-	flags = GBSerializedAudioFlagsSetCh2Hi(flags, audio->ch2.control.hi);
 	ch2Flags = GBSerializedAudioEnvelopeSetLength(ch2Flags, audio->ch2.control.length);
 	ch2Flags = GBSerializedAudioEnvelopeSetNextStep(ch2Flags, audio->ch2.envelope.nextStep);
+	ch2Flags = GBSerializedAudioEnvelopeSetDutyIndex(ch2Flags, audio->ch2.index);
 	STORE_32LE(ch2Flags, 0, &state->ch2.envelope);
-	STORE_32LE(audio->ch2Event.when - mTimingCurrentTime(audio->timing), 0, &state->ch2.nextEvent);
+	STORE_32LE(audio->ch2.lastUpdate - mTimingCurrentTime(audio->timing), 0, &state->ch2.lastUpdate);
 
 	flags = GBSerializedAudioFlagsSetCh3Readable(flags, audio->ch3.readable);
 	memcpy(state->ch3.wavebanks, audio->ch3.wavedata32, sizeof(state->ch3.wavebanks));
 	STORE_16LE(audio->ch3.length, 0, &state->ch3.length);
-	STORE_32LE(audio->ch3Event.when - mTimingCurrentTime(audio->timing), 0, &state->ch3.nextEvent);
-	STORE_32LE(audio->ch3Fade.when - mTimingCurrentTime(audio->timing), 0, &state->ch1.nextCh3Fade);
+	STORE_32LE(audio->ch3.nextUpdate - mTimingCurrentTime(audio->timing), 0, &state->ch3.nextEvent);
 
 	flags = GBSerializedAudioFlagsSetCh4Volume(flags, audio->ch4.envelope.currentVolume);
 	flags = GBSerializedAudioFlagsSetCh4Dead(flags, audio->ch4.envelope.dead);
@@ -1029,7 +1019,6 @@ void GBAudioPSGDeserialize(struct GBAudio* audio, const struct GBSerializedPSGSt
 	LOAD_32LE(sweep, 0, &state->ch1.sweep);
 	audio->ch1.envelope.currentVolume = GBSerializedAudioFlagsGetCh1Volume(flags);
 	audio->ch1.envelope.dead = GBSerializedAudioFlagsGetCh1Dead(flags);
-	audio->ch1.control.hi = GBSerializedAudioFlagsGetCh1Hi(flags);
 	audio->ch1.sweep.enable = GBSerializedAudioFlagsGetCh1SweepEnabled(flags);
 	audio->ch1.sweep.occurred = GBSerializedAudioFlagsGetCh1SweepOccurred(flags);
 	audio->ch1.sweep.time = GBSerializedAudioSweepGetTime(sweep);
@@ -1039,34 +1028,25 @@ void GBAudioPSGDeserialize(struct GBAudio* audio, const struct GBSerializedPSGSt
 	audio->ch1.control.length = GBSerializedAudioEnvelopeGetLength(ch1Flags);
 	audio->ch1.envelope.nextStep = GBSerializedAudioEnvelopeGetNextStep(ch1Flags);
 	audio->ch1.sweep.realFrequency = GBSerializedAudioEnvelopeGetFrequency(ch1Flags);
-	LOAD_32LE(when, 0, &state->ch1.nextEvent);
-	if (audio->ch1.envelope.dead < 2 && audio->playingCh1) {
-		mTimingSchedule(audio->timing, &audio->ch1Event, when);
-	}
+	audio->ch1.index = GBSerializedAudioEnvelopeGetDutyIndex(ch1Flags);
+	LOAD_32LE(audio->ch1.lastUpdate, 0, &state->ch1.lastUpdate);
+	audio->ch1.lastUpdate += mTimingCurrentTime(audio->timing);
 
 	LOAD_32LE(ch2Flags, 0, &state->ch2.envelope);
 	audio->ch2.envelope.currentVolume = GBSerializedAudioFlagsGetCh2Volume(flags);
 	audio->ch2.envelope.dead = GBSerializedAudioFlagsGetCh2Dead(flags);
-	audio->ch2.control.hi = GBSerializedAudioFlagsGetCh2Hi(flags);
 	audio->ch2.control.length = GBSerializedAudioEnvelopeGetLength(ch2Flags);
 	audio->ch2.envelope.nextStep = GBSerializedAudioEnvelopeGetNextStep(ch2Flags);
-	LOAD_32LE(when, 0, &state->ch2.nextEvent);
-	if (audio->ch2.envelope.dead < 2 && audio->playingCh2) {
-		mTimingSchedule(audio->timing, &audio->ch2Event, when);
-	}
+	audio->ch2.index = GBSerializedAudioEnvelopeGetDutyIndex(ch2Flags);
+	LOAD_32LE(audio->ch2.lastUpdate, 0, &state->ch2.lastUpdate);
+	audio->ch2.lastUpdate += mTimingCurrentTime(audio->timing);
 
 	audio->ch3.readable = GBSerializedAudioFlagsGetCh3Readable(flags);
 	// TODO: Big endian?
 	memcpy(audio->ch3.wavedata32, state->ch3.wavebanks, sizeof(audio->ch3.wavedata32));
 	LOAD_16LE(audio->ch3.length, 0, &state->ch3.length);
-	LOAD_32LE(when, 0, &state->ch3.nextEvent);
-	if (audio->playingCh3) {
-		mTimingSchedule(audio->timing, &audio->ch3Event, when);
-	}
-	LOAD_32LE(when, 0, &state->ch1.nextCh3Fade);
-	if (audio->ch3.readable && audio->style == GB_AUDIO_DMG) {
-		mTimingSchedule(audio->timing, &audio->ch3Fade, when);
-	}
+	LOAD_32LE(audio->ch3.nextUpdate, 0, &state->ch3.nextEvent);
+	audio->ch3.nextUpdate += mTimingCurrentTime(audio->timing);
 
 	LOAD_32LE(ch4Flags, 0, &state->ch4.envelope);
 	audio->ch4.envelope.currentVolume = GBSerializedAudioFlagsGetCh4Volume(flags);
@@ -1090,6 +1070,15 @@ void GBAudioPSGDeserialize(struct GBAudio* audio, const struct GBSerializedPSGSt
 
 void GBAudioSerialize(const struct GBAudio* audio, struct GBSerializedState* state) {
 	GBAudioPSGSerialize(audio, &state->audio.psg, &state->audio.flags);
+
+	size_t i;
+	for (i = 0; i < GB_MAX_SAMPLES; ++i) {
+		STORE_16LE(audio->currentSamples[i].left, 0, &state->audio2.currentSamples[i].left);
+		STORE_16LE(audio->currentSamples[i].right, 0, &state->audio2.currentSamples[i].right);
+	}
+	STORE_32LE(audio->lastSample, 0, &state->audio2.lastSample);
+	state->audio2.sampleIndex = audio->sampleIndex;
+
 	STORE_32LE(audio->capLeft, 0, &state->audio.capLeft);
 	STORE_32LE(audio->capRight, 0, &state->audio.capRight);
 	STORE_32LE(audio->sampleEvent.when - mTimingCurrentTime(audio->timing), 0, &state->audio.nextSample);
@@ -1099,6 +1088,15 @@ void GBAudioDeserialize(struct GBAudio* audio, const struct GBSerializedState* s
 	GBAudioPSGDeserialize(audio, &state->audio.psg, &state->audio.flags);
 	LOAD_32LE(audio->capLeft, 0, &state->audio.capLeft);
 	LOAD_32LE(audio->capRight, 0, &state->audio.capRight);
+
+	size_t i;
+	for (i = 0; i < GB_MAX_SAMPLES; ++i) {
+		LOAD_16LE(audio->currentSamples[i].left, 0, &state->audio2.currentSamples[i].left);
+		LOAD_16LE(audio->currentSamples[i].right, 0, &state->audio2.currentSamples[i].right);
+	}
+	LOAD_32LE(audio->lastSample, 0, &state->audio2.lastSample);
+	audio->sampleIndex = state->audio2.sampleIndex;
+
 	uint32_t when;
 	LOAD_32LE(when, 0, &state->audio.nextSample);
 	mTimingSchedule(audio->timing, &audio->sampleEvent, when);
