@@ -75,6 +75,9 @@ static void GBAInit(void* cpu, struct mCPUComponent* component) {
 	GBAMemoryInit(gba);
 
 	gba->memory.savedata.timing = &gba->timing;
+	gba->memory.savedata.vf = NULL;
+	gba->memory.savedata.realVf = NULL;
+	gba->memory.savedata.gpio = &gba->memory.hw;
 	GBASavedataInit(&gba->memory.savedata, NULL);
 
 	gba->video.p = gba;
@@ -90,14 +93,16 @@ static void GBAInit(void* cpu, struct mCPUComponent* component) {
 
 	GBAHardwareInit(&gba->memory.hw, NULL);
 
-	gba->keySource = 0;
-	gba->rotationSource = 0;
-	gba->luminanceSource = 0;
-	gba->rtcSource = 0;
-	gba->rumble = 0;
+	gba->keysActive = 0;
+	gba->keysLast = 0x400;
+	gba->rotationSource = NULL;
+	gba->luminanceSource = NULL;
+	gba->rtcSource = NULL;
+	gba->rumble = NULL;
 
-	gba->romVf = 0;
-	gba->biosVf = 0;
+	gba->romVf = NULL;
+	gba->mbVf = NULL;
+	gba->biosVf = NULL;
 
 	gba->stream = NULL;
 	gba->keyCallback = NULL;
@@ -138,7 +143,7 @@ void GBAUnloadROM(struct GBA* gba) {
 
 	if (gba->romVf) {
 #ifndef FIXED_ROM_BUFFER
-		if (gba->isPristine) {
+		if (gba->isPristine && gba->memory.rom) {
 			gba->romVf->unmap(gba->romVf, gba->memory.rom, gba->pristineRomSize);
 		}
 #endif
@@ -146,9 +151,13 @@ void GBAUnloadROM(struct GBA* gba) {
 		gba->romVf = NULL;
 	}
 	gba->memory.rom = NULL;
+	gba->memory.romSize = 0;
+	gba->memory.romMask = 0;
 	gba->isPristine = false;
 
-	gba->memory.savedata.maskWriteback = false;
+	if (!gba->memory.savedata.dirty) {
+		gba->memory.savedata.maskWriteback = false;
+	}
 	GBASavedataUnmask(&gba->memory.savedata);
 	GBASavedataDeinit(&gba->memory.savedata);
 	if (gba->memory.savedata.realVf) {
@@ -160,6 +169,7 @@ void GBAUnloadROM(struct GBA* gba) {
 
 void GBADestroy(struct GBA* gba) {
 	GBAUnloadROM(gba);
+	GBAUnloadMB(gba);
 
 	if (gba->biosVf) {
 		gba->biosVf->unmap(gba->biosVf, gba->memory.bios, SIZE_BIOS);
@@ -189,11 +199,11 @@ void GBAInterruptHandlerInit(struct ARMInterruptHandler* irqh) {
 
 void GBAReset(struct ARMCore* cpu) {
 	ARMSetPrivilegeMode(cpu, MODE_IRQ);
-	cpu->gprs[ARM_SP] = SP_BASE_IRQ;
+	cpu->gprs[ARM_SP] = GBA_SP_BASE_IRQ;
 	ARMSetPrivilegeMode(cpu, MODE_SUPERVISOR);
-	cpu->gprs[ARM_SP] = SP_BASE_SUPERVISOR;
+	cpu->gprs[ARM_SP] = GBA_SP_BASE_SUPERVISOR;
 	ARMSetPrivilegeMode(cpu, MODE_SYSTEM);
-	cpu->gprs[ARM_SP] = SP_BASE_SYSTEM;
+	cpu->gprs[ARM_SP] = GBA_SP_BASE_SYSTEM;
 
 	struct GBA* gba = (struct GBA*) cpu->master;
 	gba->memory.savedata.maskWriteback = false;
@@ -219,22 +229,24 @@ void GBAReset(struct ARMCore* cpu) {
 
 	// GB Player SIO control should not be engaged before detection, even if we already know it's GBP
 	gba->memory.hw.devices &= ~HW_GB_PLAYER;
-	if (gba->sio.drivers.normal == &gba->memory.hw.gbpDriver.d) {
+	if (gba->sio.drivers.normal == &gba->sio.gbp.d) {
 		GBASIOSetDriver(&gba->sio, NULL, SIO_NORMAL_32);
 	}
 
 	bool isELF = false;
 #ifdef USE_ELF
-	struct ELF* elf = ELFOpen(gba->romVf);
-	if (elf) {
-		isELF = true;
-		ELFClose(elf);
+	if (gba->mbVf) {
+		struct ELF* elf = ELFOpen(gba->mbVf);
+		if (elf) {
+			isELF = true;
+			ELFClose(elf);
+		}
 	}
 #endif
 
-	if (GBAIsMB(gba->romVf) && !isELF) {
-		gba->romVf->seek(gba->romVf, 0, SEEK_SET);
-		gba->romVf->read(gba->romVf, gba->memory.wram, gba->pristineRomSize);
+	if (GBAIsMB(gba->mbVf) && !isELF) {
+		gba->mbVf->seek(gba->mbVf, 0, SEEK_SET);
+		gba->mbVf->read(gba->mbVf, gba->memory.wram, SIZE_WORKING_RAM);
 	}
 
 	gba->lastJump = 0;
@@ -246,7 +258,7 @@ void GBAReset(struct ARMCore* cpu) {
 	memset(gba->debugString, 0, sizeof(gba->debugString));
 
 
-	if (gba->romVf && gba->pristineRomSize > SIZE_CART0) {
+	if (gba->romVf && gba->romVf->size(gba->romVf) > SIZE_CART0) {
 		char ident;
 		gba->romVf->seek(gba->romVf, 0xAC, SEEK_SET);
 		gba->romVf->read(gba->romVf, &ident, 1);
@@ -265,8 +277,10 @@ void GBASkipBIOS(struct GBA* gba) {
 		} else {
 			cpu->gprs[ARM_PC] = BASE_WORKING_RAM + 0xC0;
 		}
-		gba->video.vcount = 0x7D;
-		gba->memory.io[REG_VCOUNT >> 1] = 0x7D;
+		gba->video.vcount = 0x7E;
+		gba->memory.io[REG_VCOUNT >> 1] = 0x7E;
+		mTimingDeschedule(&gba->timing, &gba->video.event);
+		mTimingSchedule(&gba->timing, &gba->video.event, 117);
 		gba->memory.io[REG_POSTFLG >> 1] = 1;
 		ARMWritePC(cpu);
 	}
@@ -351,7 +365,6 @@ bool GBALoadNull(struct GBA* gba) {
 	gba->yankedRomSize = 0;
 	gba->memory.romSize = SIZE_CART0;
 	gba->memory.romMask = SIZE_CART0 - 1;
-	gba->memory.mirroring = false;
 	gba->romCrc32 = 0;
 
 	if (gba->cpu) {
@@ -362,39 +375,39 @@ bool GBALoadNull(struct GBA* gba) {
 }
 
 bool GBALoadMB(struct GBA* gba, struct VFile* vf) {
-	GBAUnloadROM(gba);
-	gba->romVf = vf;
-	gba->pristineRomSize = vf->size(vf);
+	GBAUnloadMB(gba);
+	gba->mbVf = vf;
 	vf->seek(vf, 0, SEEK_SET);
-	if (gba->pristineRomSize > SIZE_WORKING_RAM) {
-		gba->pristineRomSize = SIZE_WORKING_RAM;
-	}
-	gba->isPristine = true;
 	memset(gba->memory.wram, 0, SIZE_WORKING_RAM);
-	gba->yankedRomSize = 0;
-	gba->memory.romSize = 0;
-	gba->memory.romMask = 0;
-	gba->romCrc32 = doCrc32(gba->memory.wram, gba->pristineRomSize);
+	vf->read(vf, gba->memory.wram, SIZE_WORKING_RAM);
 	if (gba->cpu && gba->memory.activeRegion == REGION_WORKING_RAM) {
 		gba->cpu->memory.setActiveRegion(gba->cpu, gba->cpu->gprs[ARM_PC]);
 	}
 	return true;
 }
 
+void GBAUnloadMB(struct GBA* gba) {
+	if (gba->mbVf) {
+		gba->mbVf->close(gba->mbVf);
+		gba->mbVf = NULL;
+	}
+}
+
 bool GBALoadROM(struct GBA* gba, struct VFile* vf) {
 	if (!vf) {
 		return false;
 	}
 	GBAUnloadROM(gba);
 	gba->romVf = vf;
+	gba->isPristine = true;
 	gba->pristineRomSize = vf->size(vf);
 	vf->seek(vf, 0, SEEK_SET);
 	if (gba->pristineRomSize > SIZE_CART0) {
-		gba->isPristine = false;
 		char ident;
 		vf->seek(vf, 0xAC, SEEK_SET);
 		vf->read(vf, &ident, 1);
 		if (ident == 'M') {
+			gba->isPristine = false;
 			gba->memory.romSize = 0x01000000;
 #ifdef FIXED_ROM_BUFFER
 			gba->memory.rom = romBuffer;
@@ -405,8 +418,21 @@ bool GBALoadROM(struct GBA* gba, struct VFile* vf) {
 			gba->memory.rom = vf->map(vf, SIZE_CART0, MAP_READ);
 			gba->memory.romSize = SIZE_CART0;
 		}
+		gba->pristineRomSize = SIZE_CART0;
+	} else if (gba->pristineRomSize == 0x00100000) {
+		// 1 MiB ROMs (e.g. Classic NES) all appear as 4x mirrored, but not more
+		gba->isPristine = false;
+		gba->memory.romSize = 0x00400000;
+#ifdef FIXED_ROM_BUFFER
+		gba->memory.rom = romBuffer;
+#else
+		gba->memory.rom = anonymousMemoryMap(SIZE_CART0);
+#endif
+		vf->read(vf, gba->memory.rom, gba->pristineRomSize);
+		memcpy(&gba->memory.rom[0x40000], gba->memory.rom, 0x00100000);
+		memcpy(&gba->memory.rom[0x80000], gba->memory.rom, 0x00100000);
+		memcpy(&gba->memory.rom[0xC0000], gba->memory.rom, 0x00100000);
 	} else {
-		gba->isPristine = true;
 		gba->memory.rom = vf->map(vf, gba->pristineRomSize, MAP_READ);
 		gba->memory.romSize = gba->pristineRomSize;
 	}
@@ -417,8 +443,7 @@ bool GBALoadROM(struct GBA* gba, struct VFile* vf) {
 	}
 	gba->yankedRomSize = 0;
 	gba->memory.romMask = toPow2(gba->memory.romSize) - 1;
-	gba->memory.mirroring = false;
-	gba->romCrc32 = doCrc32(gba->memory.rom, gba->memory.romSize);
+	gba->romCrc32 = doCrc32(gba->memory.rom, gba->pristineRomSize);
 	if (popcount32(gba->memory.romSize) != 1) {
 		// This ROM is either a bad dump or homebrew. Emulate flash cart behavior.
 #ifndef FIXED_ROM_BUFFER
@@ -453,7 +478,7 @@ void GBAYankROM(struct GBA* gba) {
 	gba->yankedRomSize = gba->memory.romSize;
 	gba->memory.romSize = 0;
 	gba->memory.romMask = 0;
-	GBARaiseIRQ(gba, IRQ_GAMEPAK, 0);
+	GBARaiseIRQ(gba, GBA_IRQ_GAMEPAK, 0);
 }
 
 void GBALoadBIOS(struct GBA* gba, struct VFile* vf) {
@@ -542,7 +567,7 @@ void GBAHalt(struct GBA* gba) {
 }
 
 void GBAStop(struct GBA* gba) {
-	int validIrqs = (1 << IRQ_GAMEPAK) | (1 << IRQ_KEYPAD) | (1 << IRQ_SIO);
+	int validIrqs = (1 << GBA_IRQ_GAMEPAK) | (1 << GBA_IRQ_KEYPAD) | (1 << GBA_IRQ_SIO);
 	int sleep = gba->memory.io[REG_IE >> 1] & validIrqs;
 	size_t c;
 	for (c = 0; c < mCoreCallbacksListSize(&gba->coreCallbacks); ++c) {
@@ -571,20 +596,21 @@ void GBADebug(struct GBA* gba, uint16_t flags) {
 }
 
 bool GBAIsROM(struct VFile* vf) {
+	if (!vf) {
+		return false;
+	}
+
 #ifdef USE_ELF
 	struct ELF* elf = ELFOpen(vf);
 	if (elf) {
 		uint32_t entry = ELFEntry(elf);
 		bool isGBA = true;
 		isGBA = isGBA && ELFMachine(elf) == EM_ARM;
-		isGBA = isGBA && (entry == BASE_CART0 || entry == BASE_WORKING_RAM);
+		isGBA = isGBA && (entry == BASE_CART0 || entry == BASE_WORKING_RAM + 0xC0);
 		ELFClose(elf);
 		return isGBA;
 	}
 #endif
-	if (!vf) {
-		return false;
-	}
 
 	uint8_t signature[sizeof(GBA_ROM_MAGIC) + sizeof(GBA_ROM_MAGIC2)];
 	if (vf->seek(vf, GBA_ROM_MAGIC_OFFSET, SEEK_SET) < 0) {
@@ -636,7 +662,7 @@ bool GBAIsMB(struct VFile* vf) {
 #ifdef USE_ELF
 	struct ELF* elf = ELFOpen(vf);
 	if (elf) {
-		bool isMB = ELFEntry(elf) == BASE_WORKING_RAM;
+		bool isMB = ELFEntry(elf) == BASE_WORKING_RAM + 0xC0;
 		ELFClose(elf);
 		return isMB;
 	}
@@ -667,13 +693,23 @@ bool GBAIsMB(struct VFile* vf) {
 	}
 
 	uint32_t pc = GBA_MB_MAGIC_OFFSET;
+	int wramAddrs = 0;
+	int wramLoads = 0;
+	int romAddrs = 0;
+	int romLoads = 0;
 	int i;
-	for (i = 0; i < 80; ++i) {
+	for (i = 0; i < 128; ++i) {
 		if (vf->read(vf, &signature, sizeof(signature)) != sizeof(signature)) {
 			break;
 		}
 		pc += 4;
 		LOAD_32(opcode, 0, &signature);
+		if ((opcode & ~0x1FFFF) == BASE_WORKING_RAM) {
+			++wramAddrs;
+		}
+		if ((opcode & ~0x1FFFF) == BASE_CART0) {
+			++romAddrs;
+		}
 		ARMDecodeARM(opcode, &info);
 		if (info.mnemonic != ARM_MN_LDR) {
 			continue;
@@ -694,11 +730,20 @@ bool GBAIsMB(struct VFile* vf) {
 			if (vf->seek(vf, pc, SEEK_SET) < 0) {
 				break;
 			}
-			if ((immediate & ~0x7FF) == BASE_WORKING_RAM) {
-				return true;
+			if ((immediate & ~0x1FFFF) == BASE_WORKING_RAM) {
+				++wramLoads;
+			}
+			if ((immediate & ~0x1FFFF) == BASE_CART0) {
+				++romLoads;
 			}
 		}
 	}
+	if (romLoads + romAddrs >= 2) {
+		return false;
+	}
+	if (wramLoads + wramAddrs) {
+		return true;
+	}
 	// Found a libgba-linked cart...these are a bit harder to detect.
 	return false;
 }
@@ -859,7 +904,7 @@ void GBAFrameEnded(struct GBA* gba) {
 	}
 
 	if (gba->memory.hw.devices & (HW_GB_PLAYER | HW_GB_PLAYER_DETECTION)) {
-		GBAHardwarePlayerUpdate(gba);
+		GBASIOPlayerUpdate(gba);
 	}
 
 	size_t c;
@@ -875,23 +920,26 @@ void GBAFrameEnded(struct GBA* gba) {
 }
 
 void GBATestKeypadIRQ(struct GBA* gba) {
+	uint16_t keysLast = gba->keysLast;
+	uint16_t keysActive = gba->keysActive;
+
 	uint16_t keycnt = gba->memory.io[REG_KEYCNT >> 1];
 	if (!(keycnt & 0x4000)) {
 		return;
 	}
+	gba->keysLast = keysActive;
 	int isAnd = keycnt & 0x8000;
-	if (!gba->keySource) {
-		// TODO?
-		return;
-	}
 
 	keycnt &= 0x3FF;
-	uint16_t keyInput = *gba->keySource & keycnt;
-
-	if (isAnd && keycnt == keyInput) {
-		GBARaiseIRQ(gba, IRQ_KEYPAD, 0);
-	} else if (!isAnd && keyInput) {
-		GBARaiseIRQ(gba, IRQ_KEYPAD, 0);
+	if (isAnd && keycnt == (keysActive & keycnt)) {
+		if (keysLast == keysActive) {
+			return;
+		}
+		GBARaiseIRQ(gba, GBA_IRQ_KEYPAD, 0);
+	} else if (!isAnd && (keysActive & keycnt)) {
+		GBARaiseIRQ(gba, GBA_IRQ_KEYPAD, 0);
+	} else {
+		gba->keysLast = 0x400;
 	}
 }
 
