@@ -11,10 +11,13 @@
 #include "MultiplayerController.h"
 #include "Override.h"
 
+#include <QAbstractButton>
 #include <QDateTime>
+#include <QMessageBox>
 #include <QMutexLocker>
 
 #include <mgba/core/serialize.h>
+#include <mgba/core/version.h>
 #include <mgba/feature/video-logger.h>
 #ifdef M_CORE_GBA
 #include <mgba/internal/gba/gba.h>
@@ -69,6 +72,7 @@ CoreController::CoreController(mCore* core, QObject* parent)
 
 		if (controller->m_multiplayer) {
 			controller->m_multiplayer->attachGame(controller);
+			controller->updatePlayerSave();
 		}
 
 		QMetaObject::invokeMethod(controller, "started");
@@ -86,12 +90,17 @@ CoreController::CoreController(mCore* core, QObject* parent)
 		}
 
 		controller->m_resetActions.clear();
+		controller->m_frameCounter = -1;
 
 		if (!controller->m_hwaccel) {
 			context->core->setVideoBuffer(context->core, reinterpret_cast<color_t*>(controller->m_activeBuffer.data()), controller->screenDimensions().width());
 		}
 
+		QString message(tr("Reset r%1-%2 %3").arg(gitRevision).arg(QLatin1String(gitCommitShort)).arg(controller->m_crc32, 8, 16, QLatin1Char('0')));
 		QMetaObject::invokeMethod(controller, "didReset");
+		if (controller->m_showResetInfo) {
+			QMetaObject::invokeMethod(controller, "statusPosted", Q_ARG(const QString&, message));
+		}
 		controller->finishFrame();
 	};
 
@@ -135,19 +144,19 @@ CoreController::CoreController(mCore* core, QObject* parent)
 		QMetaObject::invokeMethod(controller, "unpaused");
 	};
 
-	m_threadContext.logger.d.log = [](mLogger* logger, int category, enum mLogLevel level, const char* format, va_list args) {
-		mThreadLogger* logContext = reinterpret_cast<mThreadLogger*>(logger);
-		mCoreThread* context = logContext->p;
+	m_logger.self = this;
+	m_logger.log = [](mLogger* logger, int category, enum mLogLevel level, const char* format, va_list args) {
+		CoreLogger* logContext = static_cast<CoreLogger*>(logger);
 
 		static const char* savestateMessage = "State %i saved";
 		static const char* loadstateMessage = "State %i loaded";
 		static const char* savestateFailedMessage = "State %i failed to load";
 		static int biosCat = -1;
 		static int statusCat = -1;
-		if (!context) {
+		if (!logContext) {
 			return;
 		}
-		CoreController* controller = static_cast<CoreController*>(context->userData);
+		CoreController* controller = logContext->self;
 		QString message;
 		if (biosCat < 0) {
 			biosCat = mLogCategoryById("gba.bios");
@@ -185,16 +194,17 @@ CoreController::CoreController(mCore* core, QObject* parent)
 			}
 			va_list argc;
 			va_copy(argc, args);
-			message = QString().vsprintf(format, argc);
+			message = QString::vasprintf(format, argc);
 			va_end(argc);
 			QMetaObject::invokeMethod(controller, "statusPosted", Q_ARG(const QString&, message));
 		}
-		message = QString().vsprintf(format, args);
+		message = QString::vasprintf(format, args);
 		QMetaObject::invokeMethod(controller, "logPosted", Q_ARG(int, level), Q_ARG(int, category), Q_ARG(const QString&, message));
 		if (level == mLOG_FATAL) {
 			QMetaObject::invokeMethod(controller, "crashed", Q_ARG(const QString&, message));
 		}
 	};
+	m_threadContext.logger.logger = &m_logger;
 }
 
 CoreController::~CoreController() {
@@ -277,6 +287,14 @@ void CoreController::loadConfig(ConfigController* config) {
 	m_fastForwardMute = config->getOption("fastForwardMute", -1).toInt();
 	mCoreConfigCopyValue(&m_threadContext.core->config, config->config(), "volume");
 	mCoreConfigCopyValue(&m_threadContext.core->config, config->config(), "mute");
+	m_preload = config->getOption("preload").toInt();
+
+	int playerId = m_multiplayer->playerId(this) + 1;
+	QVariant savePlayerId = config->getOption("savePlayerId");
+	if (m_multiplayer->attached() < 2 && savePlayerId.canConvert<int>()) {
+		playerId = savePlayerId.toInt();
+	}
+	mCoreConfigSetOverrideIntValue(&m_threadContext.core->config, "savePlayerId", playerId);
 
 	QSize sizeBefore = screenDimensions();
 	m_activeBuffer.resize(256 * 224 * sizeof(color_t));
@@ -292,14 +310,15 @@ void CoreController::loadConfig(ConfigController* config) {
 		updateFastForward();
 		mCoreThreadRewindParamsChanged(&m_threadContext);
 	}
-	if (sizeBefore != sizeAfter) {
 #ifdef M_CORE_GB
+	if (sizeBefore != sizeAfter) {
 		mCoreConfigSetIntValue(&m_threadContext.core->config, "sgb.borders", 0);
 		m_threadContext.core->reloadConfigOption(m_threadContext.core, "sgb.borders", nullptr);
 		mCoreConfigCopyValue(&m_threadContext.core->config, config->config(), "sgb.borders");
 		m_threadContext.core->reloadConfigOption(m_threadContext.core, "sgb.borders", nullptr);
-#endif
 	}
+	m_threadContext.core->reloadConfigOption(m_threadContext.core, "gb.pal", config->config());
+#endif
 }
 
 #ifdef USE_DEBUGGERS
@@ -405,7 +424,7 @@ void CoreController::setInputController(InputController* inputController) {
 void CoreController::setLogger(LogController* logger) {
 	disconnect(m_log);
 	m_log = logger;
-	m_threadContext.logger.d.filter = logger->filter();
+	m_logger.filter = logger->filter();
 	connect(this, &CoreController::logPosted, m_log, &LogController::postLog);
 }
 
@@ -476,8 +495,15 @@ void CoreController::setSync(bool sync) {
 	}
 }
 
+void CoreController::showResetInfo(bool enable) {
+	m_showResetInfo = enable;
+}
+
 void CoreController::setRewinding(bool rewind) {
 	if (!m_threadContext.core->opts.rewindEnable) {
+		if (rewind) {
+			emit statusPosted(tr("Rewinding not currently enabled"));
+		}
 		return;
 	}
 	if (rewind && m_multiplayer && m_multiplayer->attached() > 1) {
@@ -492,17 +518,22 @@ void CoreController::setRewinding(bool rewind) {
 }
 
 void CoreController::rewind(int states) {
-	{
-		Interrupter interrupter(this);
-		if (!states) {
-			states = INT_MAX;
-		}
-		for (int i = 0; i < states; ++i) {
-			if (!mCoreRewindRestore(&m_threadContext.impl->rewind, m_threadContext.core)) {
-				break;
-			}
+	if (!states) {
+		return;
+	}
+	if (!m_threadContext.core->opts.rewindEnable) {
+		emit statusPosted(tr("Rewinding not currently enabled"));
+	}
+	Interrupter interrupter(this);
+	if (!states) {
+		states = INT_MAX;
+	}
+	for (int i = 0; i < states; ++i) {
+		if (!mCoreRewindRestore(&m_threadContext.impl->rewind, m_threadContext.core)) {
+			break;
 		}
 	}
+	interrupter.resume();
 	emit frameAvailable();
 	emit rewound();
 }
@@ -525,11 +556,64 @@ void CoreController::forceFastForward(bool enable) {
 	emit fastForwardChanged(enable || m_fastForward);
 }
 
+void CoreController::changePlayer(int id) {
+	Interrupter interrupter(this);
+	int playerId = 0;
+	mCoreConfigGetIntValue(&m_threadContext.core->config, "savePlayerId", &playerId);
+	if (id == playerId) {
+		return;
+	}
+	interrupter.resume();
+
+	QMessageBox* resetPrompt = new QMessageBox(QMessageBox::Question, tr("Reset the game?"),
+		tr("Most games will require a reset to load the new save. Do you want to reset now?"),
+		QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
+	connect(resetPrompt, &QMessageBox::buttonClicked, this, [this, resetPrompt, id](QAbstractButton* button) {
+		Interrupter interrupter(this);
+		switch (resetPrompt->standardButton(button)) {
+		default:
+			return;
+		case QMessageBox::Yes:
+			mCoreConfigSetOverrideIntValue(&m_threadContext.core->config, "savePlayerId", id);
+			m_resetActions.append([this]() {
+				updatePlayerSave();
+			});
+			interrupter.resume();
+			reset();
+			break;
+		case QMessageBox::No:
+			mCoreConfigSetOverrideIntValue(&m_threadContext.core->config, "savePlayerId", id);
+			updatePlayerSave();
+			break;
+		}
+	});
+	resetPrompt->setAttribute(Qt::WA_DeleteOnClose);
+	resetPrompt->show();
+}
+
+void CoreController::overrideMute(bool override) {
+	m_mute = override;
+
+	Interrupter interrupter(this);
+	mCore* core = m_threadContext.core;
+	if (m_mute) {
+		core->opts.mute = true;
+	} else {
+		if (m_fastForward || m_fastForwardForced) {
+			core->opts.mute = m_fastForwardMute >= 0;
+		} else {
+			mCoreConfigGetBoolValue(&core->config, "mute", &core->opts.mute);
+		}
+	}
+	core->reloadConfigOption(core, NULL, NULL);
+}
+
 void CoreController::loadState(int slot) {
 	if (slot > 0 && slot != m_stateSlot) {
 		m_stateSlot = slot;
 		m_backupSaveState.clear();
 	}
+	mCoreThreadClearCrashed(&m_threadContext);
 	mCoreThreadRunFunction(&m_threadContext, [](mCoreThread* context) {
 		CoreController* controller = static_cast<CoreController*>(context->userData);
 		if (!controller->m_backupLoadState.isOpen()) {
@@ -549,6 +633,7 @@ void CoreController::loadState(const QString& path, int flags) {
 	if (flags != -1) {
 		m_loadStateFlags = flags;
 	}
+	mCoreThreadClearCrashed(&m_threadContext);
 	mCoreThreadRunFunction(&m_threadContext, [](mCoreThread* context) {
 		CoreController* controller = static_cast<CoreController*>(context->userData);
 		VFile* vf = VFileDevice::open(controller->m_statePath, O_RDONLY);
@@ -577,6 +662,7 @@ void CoreController::loadState(QIODevice* iodev, int flags) {
 	if (flags != -1) {
 		m_loadStateFlags = flags;
 	}
+	mCoreThreadClearCrashed(&m_threadContext);
 	mCoreThreadRunFunction(&m_threadContext, [](mCoreThread* context) {
 		CoreController* controller = static_cast<CoreController*>(context->userData);
 		VFile* vf = controller->m_stateVf;
@@ -705,7 +791,22 @@ void CoreController::loadSave(const QString& path, bool temporary) {
 			m_threadContext.core->loadSave(m_threadContext.core, vf);
 		}
 	});
-	reset();
+	if (hasStarted()) {
+		reset();
+	}
+}
+
+void CoreController::loadSave(VFile* vf, bool temporary) {
+	m_resetActions.append([this, vf, temporary]() {
+		if (temporary) {
+			m_threadContext.core->loadTemporarySave(m_threadContext.core, vf);
+		} else {
+			m_threadContext.core->loadSave(m_threadContext.core, vf);
+		}
+	});
+	if (hasStarted()) {
+		reset();
+	}
 }
 
 void CoreController::loadPatch(const QString& patchPath) {
@@ -732,7 +833,11 @@ void CoreController::replaceGame(const QString& path) {
 	QString fname = info.canonicalFilePath();
 	Interrupter interrupter(this);
 	mDirectorySetDetachBase(&m_threadContext.core->dirs);
-	mCoreLoadFile(m_threadContext.core, fname.toUtf8().constData());
+	if (m_preload) {
+		mCorePreloadFile(m_threadContext.core, fname.toUtf8().constData());
+	} else {
+		mCoreLoadFile(m_threadContext.core, fname.toUtf8().constData());
+	}
 	updateROMInfo();
 }
 
@@ -762,6 +867,7 @@ void CoreController::addKey(int key) {
 
 void CoreController::clearKey(int key) {
 	m_activeKeys &= ~(1 << key);
+	m_removedKeys |= 1 << key;
 }
 
 void CoreController::setAutofire(int key, bool enable) {
@@ -795,13 +901,41 @@ void CoreController::setFakeEpoch(const QDateTime& time) {
 	m_threadContext.core->rtc.value = time.toMSecsSinceEpoch();
 }
 
+void CoreController::setTimeOffset(qint64 offset) {
+	m_threadContext.core->rtc.override = RTC_WALLCLOCK_OFFSET;
+	m_threadContext.core->rtc.value = offset * 1000LL;
+}
+
 void CoreController::scanCard(const QString& path) {
 #ifdef M_CORE_GBA
 	QImage image(path);
 	if (image.isNull()) {
 		QFile file(path);
-		file.open(QIODevice::ReadOnly);
+		if (!file.open(QIODevice::ReadOnly)) {
+			return;
+		}
 		m_eReaderData = file.read(2912);
+
+		file.seek(0);
+		QStringList lines;
+		QDir basedir(QFileInfo(path).dir());
+
+		while (true) {
+			QByteArray line = file.readLine().trimmed();
+			if (line.isEmpty()) {
+				break;
+			}
+			QString filepath(QString::fromUtf8(line));
+			if (filepath.isEmpty() || filepath[0] == QChar('#')) {
+				continue;
+			}
+			if (QFileInfo(filepath).isRelative()) {
+				lines.append(basedir.filePath(filepath));
+			} else {
+				lines.append(filepath);
+			}
+		}
+		scanCards(lines);
 	} else if (image.size() == QSize(989, 44) || image.size() == QSize(639, 44)) {
 		const uchar* bits = image.constBits();
 		size_t size;
@@ -815,11 +949,16 @@ void CoreController::scanCard(const QString& path) {
 
 	mCoreThreadRunFunction(&m_threadContext, [](mCoreThread* thread) {
 		CoreController* controller = static_cast<CoreController*>(thread->userData);
-		GBAEReaderQueueCard(static_cast<GBA*>(thread->core->board), controller->m_eReaderData.constData(), controller->m_eReaderData.size());
+		GBACartEReaderQueueCard(static_cast<GBA*>(thread->core->board), controller->m_eReaderData.constData(), controller->m_eReaderData.size());
 	});
 #endif
 }
 
+void CoreController::scanCards(const QStringList& paths) {
+	for (const QString& path : paths) {
+		scanCard(path);
+	}
+}
 
 void CoreController::importSharkport(const QString& path) {
 #ifdef M_CORE_GBA
@@ -833,6 +972,7 @@ void CoreController::importSharkport(const QString& path) {
 	}
 	Interrupter interrupter(this);
 	GBASavedataImportSharkPort(static_cast<GBA*>(m_threadContext.core->board), vf, false);
+	GBASavedataImportGSV(static_cast<GBA*>(m_threadContext.core->board), vf, false);
 	vf->close(vf);
 #endif
 }
@@ -1019,7 +1159,10 @@ void CoreController::setFramebufferHandle(int fb) {
 }
 
 void CoreController::updateKeys() {
-	int activeKeys = m_activeKeys | updateAutofire() | m_inputController->pollEvents();
+	int polledKeys = m_inputController->pollEvents() | updateAutofire();
+	int activeKeys = m_activeKeys | polledKeys;
+	activeKeys |= m_threadContext.core->getKeys(m_threadContext.core) & ~m_removedKeys;
+	m_removedKeys = polledKeys;
 	m_threadContext.core->setKeys(m_threadContext.core, activeKeys);
 }
 
@@ -1061,12 +1204,33 @@ void CoreController::finishFrame() {
 				mCoreThreadPauseFromThread(&m_threadContext);
 			}
 		}
+		++m_frameCounter;
 	}
 	updateKeys();
 
 	QMetaObject::invokeMethod(this, "frameAvailable");
 }
 
+void CoreController::updatePlayerSave() {
+	int savePlayerId = 0;
+	mCoreConfigGetIntValue(&m_threadContext.core->config, "savePlayerId", &savePlayerId);
+	if (savePlayerId == 0 || m_multiplayer->attached() > 1) {
+		savePlayerId = m_multiplayer->playerId(this) + 1;
+	}
+
+	QString saveSuffix;
+	if (savePlayerId < 2) {
+		saveSuffix = QLatin1String(".sav");
+	} else {
+		saveSuffix = QString(".sa%1").arg(savePlayerId);
+	}
+	QByteArray saveSuffixBin(saveSuffix.toUtf8());
+	VFile* save = mDirectorySetOpenSuffix(&m_threadContext.core->dirs, m_threadContext.core->dirs.save, saveSuffixBin.constData(), O_CREAT | O_RDWR);
+	if (save) {
+		m_threadContext.core->loadSave(m_threadContext.core, save);
+	}
+}
+
 void CoreController::updateFastForward() {
 	// If we have "Fast forward" checked in the menu (m_fastForwardForced)
 	// or are holding the fast forward button (m_fastForward):
@@ -1075,35 +1239,30 @@ void CoreController::updateFastForward() {
 			m_threadContext.core->opts.volume = m_fastForwardVolume;
 		}
 		if (m_fastForwardMute >= 0) {
-			m_threadContext.core->opts.mute = m_fastForwardMute;
+			m_threadContext.core->opts.mute = m_fastForwardMute || m_mute;
 		}
+		setSync(false);
 
 		// If we aren't holding the fast forward button
 		// then use the non "(held)" ratio
 		if(!m_fastForward) {
 			if (m_fastForwardRatio > 0) {
 				m_threadContext.impl->sync.fpsTarget = m_fpsTarget * m_fastForwardRatio;
-				setSync(true);
-			}	else {
-				setSync(false);
+				m_threadContext.impl->sync.audioWait = true;
 			}
 		} else {
 			// If we are holding the fast forward button,
 			// then use the held ratio
 			if (m_fastForwardHeldRatio > 0) {
 				m_threadContext.impl->sync.fpsTarget = m_fpsTarget * m_fastForwardHeldRatio;
-				setSync(true);
-			} else {
-				setSync(false);
+				m_threadContext.impl->sync.audioWait = true;
 			}
 		}
 	} else {
 		if (!mCoreConfigGetIntValue(&m_threadContext.core->config, "volume", &m_threadContext.core->opts.volume)) {
 			m_threadContext.core->opts.volume = 0x100;
 		}
-		int fakeBool = 0;
-		mCoreConfigGetIntValue(&m_threadContext.core->config, "mute", &fakeBool);
-		m_threadContext.core->opts.mute = fakeBool;
+		mCoreConfigGetBoolValue(&m_threadContext.core->config, "mute", &m_threadContext.core->opts.mute);
 		m_threadContext.impl->sync.fpsTarget = m_fpsTarget;
 		setSync(true);
 	}
@@ -1199,3 +1358,7 @@ void CoreController::Interrupter::resume(CoreController* controller) {
 
 	mCoreThreadContinue(controller->thread());
 }
+
+bool CoreController::Interrupter::held() const {
+	return m_parent && m_parent->thread()->impl;
+}
