@@ -5,7 +5,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 #pragma once
 
-#if defined(BUILD_GL) || defined(BUILD_GLES2)
+#if defined(BUILD_GL) || defined(BUILD_GLES2) || defined(BUILD_GLES3) || defined(USE_EPOXY)
 
 #include "Display.h"
 
@@ -21,13 +21,18 @@
 #include <QHash>
 #include <QList>
 #include <QMouseEvent>
+#include <QOffscreenSurface>
 #include <QOpenGLContext>
+#include <QOpenGLShaderProgram>
+#include <QOpenGLVertexArrayObject>
+#include <QOpenGLWidget>
 #include <QPainter>
 #include <QQueue>
 #include <QThread>
 #include <QTimer>
 
 #include <array>
+#include <memory>
 
 #include "CoreController.h"
 #include "VideoProxy.h"
@@ -35,11 +40,40 @@
 #include "platform/video-backend.h"
 
 class QOpenGLPaintDevice;
+class QOpenGLWidget;
 
 uint qHash(const QSurfaceFormat&, uint seed = 0);
 
 namespace QGBA {
 
+class mGLWidget : public QOpenGLWidget {
+Q_OBJECT
+
+public:
+	mGLWidget(QWidget* parent = nullptr);
+
+	void setTex(GLuint tex) { m_tex = tex; }
+	void setVBO(GLuint vbo) { m_vbo = vbo; }
+	bool finalizeVAO();
+	void reset();
+
+protected:
+	void initializeGL() override;
+	void paintGL() override;
+
+private:
+	GLuint m_tex;
+	GLuint m_vbo;
+
+	bool m_vaoDone = false;
+	std::unique_ptr<QOpenGLVertexArrayObject> m_vao;
+	std::unique_ptr<QOpenGLShaderProgram> m_program;
+	GLuint m_positionLocation;
+
+	QTimer m_refresh;
+	int m_refreshResidue = 0;
+};
+
 class PainterGL;
 class DisplayGL : public Display {
 Q_OBJECT
@@ -66,6 +100,7 @@ public slots:
 	void lockIntegerScaling(bool lock) override;
 	void interframeBlending(bool enable) override;
 	void showOSDMessages(bool enable) override;
+	void showFrameCounter(bool enable) override;
 	void filter(bool filter) override;
 	void framePosted() override;
 	void setShaders(struct VDir*) override;
@@ -79,6 +114,7 @@ protected:
 
 private:
 	void resizePainter();
+	bool shouldDisableUpdates();
 
 	static QHash<QSurfaceFormat, bool> s_supports;
 
@@ -87,13 +123,14 @@ private:
 	std::unique_ptr<PainterGL> m_painter;
 	QThread m_drawThread;
 	std::shared_ptr<CoreController> m_context;
+	mGLWidget* m_gl;
 };
 
 class PainterGL : public QObject {
 Q_OBJECT
 
 public:
-	PainterGL(QWindow* surface, const QSurfaceFormat& format);
+	PainterGL(QWindow* surface, mGLWidget* widget, const QSurfaceFormat& format);
 	~PainterGL();
 
 	void setThread(QThread*);
@@ -101,6 +138,8 @@ public:
 	void setMessagePainter(MessagePainter*);
 	void enqueue(const uint32_t* backing);
 
+	void stop();
+
 	bool supportsShaders() const { return m_supportsShaders; }
 	int glTex();
 
@@ -114,7 +153,6 @@ public slots:
 	void forceDraw();
 	void draw();
 	void start();
-	void stop();
 	void pause();
 	void unpause();
 	void resize(const QSize& size);
@@ -122,6 +160,7 @@ public slots:
 	void lockIntegerScaling(bool lock);
 	void interframeBlending(bool enable);
 	void showOSD(bool enable);
+	void showFrameCounter(bool enable);
 	void filter(bool filter);
 	void resizeContext();
 
@@ -132,6 +171,9 @@ public slots:
 signals:
 	void started();
 
+private slots:
+	void doStop();
+
 private:
 	void makeCurrent();
 	void performDraw();
@@ -144,10 +186,14 @@ private:
 	uint32_t* m_buffer = nullptr;
 	QPainter m_painter;
 	QMutex m_mutex;
-	QWindow* m_surface;
+	QWindow* m_window;
+	QSurface* m_surface;
 	QSurfaceFormat m_format;
-	std::unique_ptr<QOpenGLPaintDevice> m_window;
+	std::unique_ptr<QOpenGLPaintDevice> m_paintDev;
 	std::unique_ptr<QOpenGLContext> m_gl;
+	int m_finalTexIdx = 0;
+	GLuint m_finalTex[2];
+	mGLWidget* m_widget;
 	bool m_active = false;
 	bool m_started = false;
 	QTimer m_drawTimer;
@@ -155,6 +201,7 @@ private:
 	CoreController::Interrupter m_interrupter;
 	bool m_supportsShaders;
 	bool m_showOSD;
+	bool m_showFrameCounter;
 	VideoShader m_shader{};
 	VideoBackend* m_backend = nullptr;
 	QSize m_size;
