@@ -1,4 +1,4 @@
-/* Copyright (c) 2013-2014 Jeffrey Pfau
+/* Copyright (c) 2013-2021 Jeffrey Pfau
  *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -6,6 +6,7 @@
 #include "SettingsView.h"
 
 #include "AudioProcessor.h"
+#include "CheckBoxDelegate.h"
 #include "ConfigController.h"
 #include "Display.h"
 #include "GBAApp.h"
@@ -17,6 +18,7 @@
 
 #ifdef M_CORE_GB
 #include "GameBoy.h"
+#include <mgba/internal/gb/overrides.h>
 #endif
 
 #include <mgba/core/serialize.h>
@@ -33,15 +35,17 @@ SettingsView::SettingsView(ConfigController* controller, InputController* inputC
 	m_ui.setupUi(this);
 
 	m_pageIndex[Page::AV] = 0;
-	m_pageIndex[Page::INTERFACE] = 1;
-	m_pageIndex[Page::EMULATION] = 2;
-	m_pageIndex[Page::ENHANCEMENTS] = 3;
-	m_pageIndex[Page::BIOS] = 4;
-	m_pageIndex[Page::PATHS] = 5;
-	m_pageIndex[Page::LOGGING] = 6;
+	m_pageIndex[Page::GAMEPLAY] = 1;
+	m_pageIndex[Page::INTERFACE] = 2;
+	m_pageIndex[Page::UPDATE] = 3;
+	m_pageIndex[Page::EMULATION] = 4;
+	m_pageIndex[Page::ENHANCEMENTS] = 5;
+	m_pageIndex[Page::BIOS] = 6;
+	m_pageIndex[Page::PATHS] = 7;
+	m_pageIndex[Page::LOGGING] = 8;
 
 #ifdef M_CORE_GB
-	m_pageIndex[Page::GB] = 7;
+	m_pageIndex[Page::GB] = 9;
 
 	for (auto model : GameBoy::modelList()) {
 		m_ui.gbModel->addItem(GameBoy::modelName(model), model);
@@ -174,6 +178,38 @@ SettingsView::SettingsView(ConfigController* controller, InputController* inputC
 	}
 #endif
 
+	ApplicationUpdater* updater = GBAApp::app()->updater();
+	m_ui.currentChannel->setText(ApplicationUpdater::readableChannel());
+	m_ui.currentVersion->setText(ApplicationUpdater::currentVersion());
+	QDateTime lastCheck = updater->lastCheck();
+	if (!lastCheck.isNull()) {
+		m_ui.lastChecked->setText(lastCheck.toLocalTime().toString());
+	}
+	connect(m_ui.checkUpdate, &QAbstractButton::pressed, updater, &ApplicationUpdater::checkUpdate);
+	connect(updater, &ApplicationUpdater::updateAvailable, this, [this, updater](bool hasUpdate) {
+		updateChecked();
+		if (hasUpdate) {
+			m_ui.availVersion->setText(updater->updateInfo());
+		}
+	});
+	for (const QString& channel : ApplicationUpdater::listChannels()) {
+		m_ui.updateChannel->addItem(ApplicationUpdater::readableChannel(channel), channel);
+		if (channel == ApplicationUpdater::currentChannel()) {
+			m_ui.updateChannel->setCurrentIndex(m_ui.updateChannel->count() - 1);
+		}
+	}
+	connect(m_ui.updateChannel, static_cast<void (QComboBox::*)(int)>(&QComboBox::currentIndexChanged), this, [this, updater](int) {
+		QString channel = m_ui.updateChannel->currentData().toString();
+		updater->setChannel(channel);
+		auto updates = updater->listUpdates();
+		if (updates.contains(channel)) {
+			m_ui.availVersion->setText(updates[channel]);
+		} else {
+			m_ui.availVersion->setText(tr("None"));
+		}
+	});
+	m_ui.availVersion->setText(updater->updateInfo());
+
 	// TODO: Move to reloadConfig()
 	QVariant cameraDriver = m_controller->getQtOption("cameraDriver");
 	m_ui.cameraDriver->addItem(tr("None (Still Image)"), static_cast<int>(InputController::CameraDriver::NONE));
@@ -258,6 +294,20 @@ SettingsView::SettingsView(ConfigController* controller, InputController* inputC
 			m_gbColors[colorId] = color.rgb();
 		});
 	}
+
+	const GBColorPreset* colorPresets;
+	size_t nPresets = GBColorPresetList(&colorPresets);
+	for (size_t i = 0; i < nPresets; ++i) {
+		m_ui.colorPreset->addItem(QString(colorPresets[i].name));
+	}
+	connect(m_ui.colorPreset, static_cast<void (QComboBox::*)(int)>(&QComboBox::currentIndexChanged), this, [this, colorPresets](int n) {
+		const GBColorPreset* preset = &colorPresets[n];
+		for (int colorId = 0; colorId < 12; ++colorId) {
+			uint32_t color = preset->colors[colorId] | 0xFF000000;
+			m_colorPickers[colorId].setColor(color);
+			m_gbColors[colorId] = color;
+		}
+	});
 #else
 	m_ui.gbBiosBrowse->hide();
 	m_ui.gbcBiosBrowse->hide();
@@ -289,7 +339,8 @@ SettingsView::SettingsView(ConfigController* controller, InputController* inputC
 		}
 	});
 
-	m_ui.languages->setItemData(0, QLocale("en"));
+	QLocale englishLocale("en");
+	m_ui.languages->addItem(englishLocale.nativeLanguageName(), englishLocale);
 	QDir ts(":/translations/");
 	for (auto name : ts.entryList()) {
 		if (!name.endsWith(".qm") || !name.startsWith(binaryName)) {
@@ -306,9 +357,11 @@ SettingsView::SettingsView(ConfigController* controller, InputController* inputC
 	}
 
 	m_ui.loggingView->setModel(&m_logModel);
+	m_ui.loggingView->setItemDelegate(new CheckBoxDelegate(m_ui.loggingView));
 	m_ui.loggingView->setHorizontalHeader(new RotatedHeaderView(Qt::Horizontal));
 	m_ui.loggingView->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
 	m_ui.loggingView->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
+	connect(m_ui.loggingView, SIGNAL(clicked(QModelIndex)), m_ui.loggingView, SLOT(setCurrentIndex(QModelIndex)));
 
 	connect(m_ui.logFileBrowse, &QAbstractButton::pressed, [this] () {
 		QString path = GBAApp::app()->getSaveFileName(this, "Select log file");
@@ -317,6 +370,12 @@ SettingsView::SettingsView(ConfigController* controller, InputController* inputC
 		}
 	});
 
+	m_checkTimer.setInterval(60);
+	m_checkTimer.setSingleShot(false);
+	connect(&m_checkTimer, &QTimer::timeout, this, &SettingsView::updateChecked);
+	m_checkTimer.start();
+	updateChecked();
+
 	ShortcutView* shortcutView = new ShortcutView();
 	shortcutView->setController(shortcutController);
 	shortcutView->setInputController(inputController);
@@ -384,7 +443,6 @@ void SettingsView::updateConfig() {
 	saveSetting("gbc.bios", m_ui.gbcBios);
 	saveSetting("sgb.bios", m_ui.sgbBios);
 	saveSetting("sgb.borders", m_ui.sgbBorders);
-	saveSetting("useCgbColors", m_ui.useCgbColors);
 	saveSetting("useBios", m_ui.useBios);
 	saveSetting("skipBios", m_ui.skipBios);
 	saveSetting("sampleRate", m_ui.sampleRate);
@@ -396,6 +454,8 @@ void SettingsView::updateConfig() {
 	saveSetting("lockIntegerScaling", m_ui.lockIntegerScaling);
 	saveSetting("interframeBlending", m_ui.interframeBlending);
 	saveSetting("showOSD", m_ui.showOSD);
+	saveSetting("showFrameCounter", m_ui.showFrameCounter);
+	saveSetting("showResetInfo", m_ui.showResetInfo);
 	saveSetting("volume", m_ui.volume);
 	saveSetting("mute", m_ui.mute);
 	saveSetting("fastForwardVolume", m_ui.volumeFf);
@@ -407,6 +467,8 @@ void SettingsView::updateConfig() {
 	saveSetting("suspendScreensaver", m_ui.suspendScreensaver);
 	saveSetting("pauseOnFocusLost", m_ui.pauseOnFocusLost);
 	saveSetting("pauseOnMinimize", m_ui.pauseOnMinimize);
+	saveSetting("muteOnFocusLost", m_ui.muteOnFocusLost);
+	saveSetting("muteOnMinimize", m_ui.muteOnMinimize);
 	saveSetting("savegamePath", m_ui.savegamePath);
 	saveSetting("savestatePath", m_ui.savestatePath);
 	saveSetting("screenshotPath", m_ui.screenshotPath);
@@ -425,11 +487,12 @@ void SettingsView::updateConfig() {
 	saveSetting("logToStdout", m_ui.logToStdout);
 	saveSetting("logFile", m_ui.logFile);
 	saveSetting("useDiscordPresence", m_ui.useDiscordPresence);
-	saveSetting("gba.audioHle", m_ui.audioHle);
 	saveSetting("dynamicTitle", m_ui.dynamicTitle);
 	saveSetting("videoScale", m_ui.videoScale);
 	saveSetting("gba.forceGbp", m_ui.forceGbp);
 	saveSetting("vbaBugCompat", m_ui.vbaBugCompat);
+	saveSetting("updateAutoCheck", m_ui.updateAutoCheck);
+	saveSetting("showFilenameInLibrary", m_ui.showFilenameInLibrary);
 
 	if (m_ui.audioBufferSize->currentText().toInt() > 8192) {
 		m_ui.audioBufferSize->setCurrentText("8192");
@@ -518,6 +581,20 @@ void SettingsView::updateConfig() {
 		emit languageChanged();
 	}
 
+	bool oldAudioHle = m_controller->getOption("gba.audioHle", "0") != "0";
+	if (oldAudioHle != m_ui.audioHle->isChecked()) {
+		saveSetting("gba.audioHle", m_ui.audioHle);
+		emit audioHleChanged();
+	}
+
+	if (m_ui.multiplayerAudioAll->isChecked()) {
+		m_controller->setQtOption("multiplayerAudio", "all");
+	} else if (m_ui.multiplayerAudio1->isChecked()) {
+		m_controller->setQtOption("multiplayerAudio", "p1");
+	} else if (m_ui.multiplayerAudioActive->isChecked()) {
+		m_controller->setQtOption("multiplayerAudio", "active");
+	}
+
 	int hwaccelVideo = m_controller->getOption("hwaccelVideo").toInt();
 	saveSetting("hwaccelVideo", m_ui.hwaccelVideo->currentIndex());
 	if (hwaccelVideo != m_ui.hwaccelVideo->currentIndex()) {
@@ -563,6 +640,18 @@ void SettingsView::updateConfig() {
 		m_controller->setOption(color.toUtf8().constData(), m_gbColors[colorId] & ~0xFF000000);
 
 	}
+
+	int gbColors = GB_COLORS_CGB;
+	if (m_ui.gbColor->isChecked()) {
+		gbColors = GB_COLORS_NONE;
+	} else if (m_ui.cgbColor->isChecked()) {
+		gbColors = GB_COLORS_CGB;
+	} else if (m_ui.sgbColor->isChecked()) {
+		gbColors = GB_COLORS_SGB;
+	} else if (m_ui.scgbColor->isChecked()) {
+		gbColors = GB_COLORS_SGB_CGB_FALLBACK;
+	}
+	saveSetting("gb.colors", gbColors);
 #endif
 
 	m_controller->write();
@@ -571,14 +660,13 @@ void SettingsView::updateConfig() {
 	emit biosLoaded(mPLATFORM_GBA, m_ui.gbaBios->text());
 }
 
-void SettingsView::reloadConfig() {	
+void SettingsView::reloadConfig() {
 	loadSetting("bios", m_ui.gbaBios);
 	loadSetting("gba.bios", m_ui.gbaBios);
 	loadSetting("gb.bios", m_ui.gbBios);
 	loadSetting("gbc.bios", m_ui.gbcBios);
 	loadSetting("sgb.bios", m_ui.sgbBios);
 	loadSetting("sgb.borders", m_ui.sgbBorders, true);
-	loadSetting("useCgbColors", m_ui.useCgbColors, true);
 	loadSetting("useBios", m_ui.useBios);
 	loadSetting("skipBios", m_ui.skipBios);
 	loadSetting("audioBuffers", m_ui.audioBufferSize);
@@ -592,6 +680,8 @@ void SettingsView::reloadConfig() {
 	loadSetting("lockIntegerScaling", m_ui.lockIntegerScaling);
 	loadSetting("interframeBlending", m_ui.interframeBlending);
 	loadSetting("showOSD", m_ui.showOSD, true);
+	loadSetting("showFrameCounter", m_ui.showFrameCounter);
+	loadSetting("showResetInfo", m_ui.showResetInfo);
 	loadSetting("volume", m_ui.volume, 0x100);
 	loadSetting("mute", m_ui.mute, false);
 	loadSetting("fastForwardVolume", m_ui.volumeFf, m_ui.volume->value());
@@ -624,6 +714,8 @@ void SettingsView::reloadConfig() {
 	loadSetting("dynamicTitle", m_ui.dynamicTitle, true);
 	loadSetting("gba.forceGbp", m_ui.forceGbp);
 	loadSetting("vbaBugCompat", m_ui.vbaBugCompat, true);
+	loadSetting("updateAutoCheck", m_ui.updateAutoCheck);
+	loadSetting("showFilenameInLibrary", m_ui.showFilenameInLibrary);
 
 	m_ui.libraryStyle->setCurrentIndex(loadSetting("libraryStyle").toInt());
 
@@ -710,6 +802,22 @@ void SettingsView::reloadConfig() {
 		int index = m_ui.cgbSgbModel->findData(model);
 		m_ui.cgbSgbModel->setCurrentIndex(index >= 0 ? index : 0);
 	}
+
+	switch (m_controller->getOption("gb.colors", m_controller->getOption("useCgbColors", true).toInt()).toInt()) {
+	case GB_COLORS_NONE:
+		m_ui.gbColor->setChecked(true);
+		break;
+	default:
+	case GB_COLORS_CGB:
+		m_ui.cgbColor->setChecked(true);
+		break;
+	case GB_COLORS_SGB:
+		m_ui.sgbColor->setChecked(true);
+		break;
+	case GB_COLORS_SGB_CGB_FALLBACK:
+		m_ui.scgbColor->setChecked(true);
+		break;
+	}
 #endif
 
 	int hwaccelVideo = m_controller->getOption("hwaccelVideo", 0).toInt();
@@ -719,6 +827,40 @@ void SettingsView::reloadConfig() {
 		m_ui.videoScaleSize->setText(tr("(%1×%2)").arg(GBA_VIDEO_HORIZONTAL_PIXELS * value).arg(GBA_VIDEO_VERTICAL_PIXELS * value));
 	});
 	loadSetting("videoScale", m_ui.videoScale, 1);
+
+	QString multiplayerAudio = m_controller->getQtOption("multiplayerAudio").toString();
+	if (multiplayerAudio == QLatin1String("p1")) {
+		m_ui.multiplayerAudio1->setChecked(true);
+	} else if (multiplayerAudio == QLatin1String("active")) {
+		m_ui.multiplayerAudioActive->setChecked(true);
+	} else {
+		m_ui.multiplayerAudioAll->setChecked(true);
+	}
+}
+
+void SettingsView::updateChecked() {
+	QDateTime now(QDateTime::currentDateTimeUtc());
+	QDateTime lastCheck(GBAApp::app()->updater()->lastCheck());
+	if (!lastCheck.isValid()) {
+		m_ui.lastChecked->setText(tr("Never"));
+		return;
+	}
+	qint64 ago = GBAApp::app()->updater()->lastCheck().secsTo(now);
+	if (ago < 60) {
+		m_ui.lastChecked->setText(tr("Just now"));
+		return;
+	}
+	if (ago < 3600) {
+		m_ui.lastChecked->setText(tr("Less than an hour ago"));
+		return;
+	}
+	ago /= 3600;
+	if (ago < 24) {
+		m_ui.lastChecked->setText(tr("%n hour(s) ago", nullptr, ago));
+		return;
+	}
+	ago /= 24;
+	m_ui.lastChecked->setText(tr("%n day(s) ago", nullptr, ago));
 }
 
 void SettingsView::addPage(const QString& name, QWidget* view, Page index) {
